'''Gui for scattering the linear array of hard spheres'''
# import importlib as imp
import os,time,copy,easygui,matplotlib, pandas as pd, numpy as np, matplotlib.pyplot as plt
from matplotlib.widgets import CheckButtons
from . import displayStandards as dsp   #;imp.reload(dsp)
from . import glob_colors as colors     #;imp.reload(colors)
from . import qdot_sphere_array as qsa  #;imp.reload(qsa)
from . import hard_sphere_array as hsa  #;imp.reload(hsa)
from . import gui_config as config      #;imp.reload(config)


wm,hm = 0.05,0.07
props = {'lw':2,'fonts':{'title':15,'lab':13,'leg':15,'tick':10},'opt':''}
ax_props = [
    {'axPos':[0.0+wm,0.5+hm,0.5-2*wm,0.5-1.5*hm],},
    # {'axPos':[0.0+wm,0.5+hm,0.5-2*wm,0.5-1.5*hm],'labs':[r'Normalized radius $ka$',r'Normalized Cross section $\sigma/ka^2$']},
    {'axPos':[0.0+wm,0.0+hm,0.5-2*wm,0.5-1.5*hm],'labs':[r'$\Re$',r'$\Im$'  ]},
    {'axPos':[0.5+wm,0.5+hm,0.5-2*wm,0.5-1.5*hm],},#'xylims':['x',0,180]        },
    {'axPos':[0.5+wm,0.0+hm,0.5-2*wm,0.5-1.5*hm],'labs':['$y$','$z$'        ]},
]
for d in ax_props:d.update(props)
labs0 = {'kdka' :r'Normalized distance $kd/ka$',
         'kd'   :r'Normalized distance $kd$',
         'ka'   :r'Normalized radius $ka$',
         'sigka':r'Normalized Cross section $\sigma/ka^2$',
         'sigma':r'Cross section $\sigma$',}
clab0 = {'kdka':r'$kd/ka$','kd':'$kd$','ka':'$ka$','sigka':r'$\sigma/ka^2$','sigma':r'$\sigma$'}
btns_pos={'open':[0.01,0.99],'axis':[0.035,0.99], 'settings':[0.057,0.99],'help':[0.093,0.99]}

class GUI_handler():
    def __init__(self,df_name,Qdot=1,cols=['kd','sigka','ka']):
        self.fig=plt.figure()
        manager = plt.get_current_fig_manager()
        manager.window.showMaximized()
        self.ax = [self.fig.add_subplot(position=ax_props[i]['axPos']) for i in range(len(ax_props))]
        self.cid = self.fig.canvas.mpl_connect('key_press_event', self)
        self.cid = self.fig.canvas.mpl_connect('button_press_event', self)
        self.cid = self.fig.canvas.mpl_connect('motion_notify_event', self)

        self.bg = self.fig.canvas.copy_from_bbox(self.ax[0].bbox)
        self.init(df_name,Qdot,cols,v=1)
        plt.show(block=True )
        # plt.show(block=False)
        # self.refresh()

    def __call__(self, event):
        #### Mouse events
        if type(event)==matplotlib.backend_bases.MouseEvent:
            #### axes
            if event.inaxes==self.ax[0]:
                xydat  = np.array([event.xdata, event.ydata])             #; print(xydat)
                xydat  = self._normalize_dat(xydat)                       #; print(xydat)
                id0    = np.linalg.norm(self.dats-xydat,axis=1).argmin()  #; print(id0)
                xydat0 = self.df.loc[id0][self.cols0].values              #; print(xydat0)
                dist   = np.linalg.norm(self._normalize_dat(xydat0)-xydat)#; print(dist)
                self.update(button=event.button==1,hover=dist<0.1,id0=id0)#; print('event');print('   xydat:',xydat);print('   xydat0:',xydat0, 'id0',id0);print('   dist:',dist)
            #### buttons
            if event.button==1:
                for i,btn in enumerate(self.btns):
                    if btn.contains(event)[0]:self.btns_cb[i]()
            self._plot_btn(int(self.plot_btn.contains(event)[0]),int(event.button==1))
        #### keys
        elif type(event)==matplotlib.backend_bases.KeyEvent:
            if event.key=='f1': self.help()
            if event.key=='f5': self.refresh()
            if event.key==' ' : self.refresh()
            if event.key=='enter': self.change_settings()
            if event.key in ['up','right','left','down']:
                self.arrowActionsParam(event.key)

    ################################################################################################
    # callbacks
    ################################################################################################
    def _ckbtns(self,label):
        '''checkbox callback'''
        if label=='not\ncoupled':
            self.uncoupled = not self.uncoupled
            if self.uncoupled:
                self.s0 = copy.copy(self.s)
                self.s0.solve(copt=0)
        else:
            l = {'re':0,'im':1,r'$\phi$':2,'$||$':3,'$||^2$':4}[label]
            self.fopts[l] = not self.fopts[l]
        self.ax[2].cla()
        self.plot_f()
        self.refresh()

    def _plot_btn(self,hover,btn):
        plotBtnTxt = [[['b','click to plot nf'],['b','click to plot nf']],
                      [['g','click to plot nf'],['r','... plotting ...']]]
        c,txt = plotBtnTxt[hover][btn]
        self.plot_btn.set_backgroundcolor(c)
        self.plot_btn.set_text(txt.ljust(20))
        self.fig.draw_artist(self.plot_btn)
        self.fig.canvas.update()
        self.fig.canvas.flush_events()
        if hover and btn :
            self.plot_fields(field=1)
            c,txt = plotBtnTxt[0][0]
            self.plot_btn.set_backgroundcolor(c)
            self.plot_btn.set_text(txt.ljust(20))
            self.refresh()

    def arrowActionsParam(self,key):
        #### move up or down across the colored curves
        if key == 'up' or key=='down':
            d = self.df.loc[self.id0] #active data point
            idcb0 = np.abs(d[self.cb0col]-self.cb0vals).argmin()      #;print(idcb0)
            if   key == 'up'  :ik = min(self.cb0vals.size-1,idcb0+1)  #;print(ik)
            elif key == 'down':ik = max(0,idcb0-1)                    #;print(ik)
            df0 = self.df[self.df[self.cb0col]==self.cb0vals[ik]]     #;print(ik)
            id0 = np.abs(df0[self.cols0[0]]-d[self.cols0[0]]).argmin()#;print(id0)
            id0 = df0.iloc[id0].name
        #### move left or right along the curves
        elif key == 'right':
            id0 = min(self.id0+1,self.df.shape[0]-1)
        elif key == 'left':
            id0 = max(self.id0-1,0)
        if not id0==self.id0:
            self.update(button=1,hover=0,id0=id0)
        else:
            print('no data point at that value')

    def open(self):
        # datapath = os.path.realpath(os.path.dirname(__file__)+'/../../tests/scattering/data')+'/*.pkl' #; print(datapath)
        datapath = os.path.realpath(os.path.dirname(self.df_name))+'/*.pkl' #; print(datapath)
        df_name = easygui.fileopenbox(msg='open datafile', title='open', default=datapath,
            filetypes=["*.pkl"], multiple=False)
        if df_name:
            print(df_name)
            Qdot = 'qdot' in df_name
            self.load_data(df_name,Qdot,self.cols0+[self.cb0col])
            self.init_plots()

    def set_ax0(self):
        fieldNames = ['xlab','ylab','clab']
        fieldValues = [self.cols0[0],self.cols0[1],self.cb0col]
        msg = "change axis labels. Available values are :\n %s" %list(labs0.keys())
        fieldValues = config.multenterbox(msg,"ax0",fieldValues,fieldNames)
        if fieldValues:
            xlab,ylab,self.cb0col = fieldValues
            self.cols0 = [xlab,ylab]
            self.set_dats()
            self.ax[0].cla()
            self.h0  = self.ax[0].plot(0,0,'s',color='k' ,markersize=12,animated=True)[0]
            self.h1  = self.ax[0].plot(0,0,'o',color='g' ,markersize=10,animated=True)[0]
            self.update_cb0()
            self.plot_s()
            self.refresh()

    def change_settings(self):
        fieldNames = ['opts3', 'npts','range nY','range nZ', 'cmap(sigma)','cmap(l)','cmap(near field)','cmax ']
        fieldValues = [self.opts3,str(self.npts),str(self.r[0]),'%.2f,%.2f' %(self.r[1],self.r[2]), self.cmS,self.cmF,self.cmL, str(self.cmax)]
        fieldValues = config.multenterbox("Change settings","settings", fieldValues,fieldNames)
        if fieldValues:
            self.opts3,npts,ry,rz,self.cmS,self.cmF,self.cmL,cmax = fieldValues
            self.npts = max(10,min(int(npts),1000))
            self.r = np.array([ry]+rz.split(','),dtype=float)
            # self.cmax = max(0,int(cmax))
            # self.cbs[3].set_caxis([-self.cmax,self.cmax])
            self.refresh()

    def help(self):
        print(colors.green+'Commands : '+colors.black)
        print(colors.yellow+config.help_cmd+colors.black)
        easygui.msgbox(config.help_msg+config.help_cmd,title='help') #text='More info at : \n'+url)

    ##############################################################################################################
    ##### displays
    ##############################################################################################################
    ##ax0
    def plot_s(self):
        # print("...plotting cross sections")
        plts = []
        self.csS = dsp.getCs(self.cmS,self.cb0vals.size+1)
        for i,val in enumerate(self.cb0vals):
            dfa = self.df[self.df[self.cb0col]==val]
            plts += [[dfa[self.cols0[0]], dfa[self.cols0[1]],[self.csS[i],'-o' ],'']] #,'$ka=%.1f$' %ka]]
        xM,yM = self.df[self.cols0].max()
        xm,ym = self.df[self.cols0].min()
        labs = labs0[self.cols0[0]],labs0[self.cols0[1]]
        dsp.stddisp(plts,ax=self.ax[0],labs=labs,xylims=[0.99*xm,1.02*xM,0.99*ym,1.02*yM],**ax_props[0])#,pargs={'picker':1,'pickradius':1})

    ##ax1
    def plot_cp(self):
        # print('...plotting coefficients...')
        nmax = self.s.nmax
        idp = self.ip*nmax
        idl = np.arange(nmax)
        val0 = self.df.loc[self.id0][self.cb0col]
        coeff = ['Cp','bp'][self.Qdot]

        #### all coefficients on same curve as active point
        self.csL = dsp.getCs(self.cmL,nmax)
        df = self.df[self.df[self.cb0col]==val0] #;print(df)
        cp = np.vstack(df[coeff].values)         #;print(cp.shape) cause df.Cp.values=[Cp0,Cp1,...] is a list
        mm = np.abs(cp).max()
        plts  = [[ cp[:,idp+l].real ,cp[:,idp+l].imag,[self.csL[l],'-o' ]] for i,l in enumerate(idl)]
        if self.uncoupled:
            cp0 = self.s0.__getattribute__(coeff)
            plts+=[ [cp0[l].real,cp0[l].imag,[self.csL[l],'s'],''] for l in idl ]

        #### active point coefficients
        cpi = self.df.loc[self.id0][coeff][self.ip*nmax+idl] #;print(cpi)
        plts+=[[cpi.real,cpi.imag,'go','']]
        tle = '$b_{pl}$ for $%s=%g$ and varying $%s$ ' %(self.cb0col,val0,self.cols0[0])
        dsp.stddisp(plts,ax=self.ax[1],xylims=[-mm,mm,-mm,mm],title=tle,**ax_props[1])
    ##ax2
    def plot_f(self):
        # print('...plotting scatering amplitude...')
        fopts = ''.join(np.array(['r','i','a','m','2'])[self.fopts])
        legElt={}
        if self.uncoupled:
            self.s0.show_ff(npts=361,fopts=fopts,ax=self.ax[2],leg=0,title=None,**ax_props[2])
            legElt={'uncoupled':'k--'}
        self.s.show_ff(npts=361,fopts=fopts,ax=self.ax[2],legElt=legElt,**ax_props[2])
    ##ax3
    def plot_fields(self, field=1):
        ka,kd = self.s.ka,self.s.kd
        em = 0.05
        if self.r[0]==-1:
            ry = 2+kd/ka+self.r[1]+self.r[2]
        else:
            ry = 1+self.r[0]
        r = np.array([1e-3,ry*ka,-ka*(1+self.r[1]),kd+ka*(1+self.r[2])])
        xylims = [0,r[1]-em,r[2],r[3]-em]
        if field:
            # print('...plotting near fields...')
            self.s.show_f(cart_opt=True,npts=self.npts,r=r,opts=self.opts3,def_args=False,fz=np.real,
                xylims=xylims,cmap=self.cmF,imOpt='c',caxis=[-self.cmax,self.cmax],cb_pos=[0.955,hm,0.015,0.5-1.5*hm],fig=self.fig,
                cb=self.cbs[3],ax=self.ax[3],**ax_props[3])
        else:
            t = np.linspace(-np.pi/2,np.pi/2,100)
            ct,st = np.cos(t), np.sin(t)
            plts = [ [ka*ct, kd*p+ka*st,'k-',''] for p in range(self.s.N)]
            dsp.stddisp(plts,xylims=xylims,ax=self.ax[3],**ax_props[3])

    ##############################################################################################################
    #### initialization
    ##############################################################################################################
    def init(self,df_name,Qdot,cols,v=0):
        print(colors.green+'Initialization'+colors.black)
        t=time.time()
        self.init_params()          #;print("init_params %.3f" %(time.time()-t));t=time.time()
        self.load_data(df_name,Qdot,cols)#;print("load %.3f" %(time.time()-t));t=time.time()
        self.init_widgets()                ;print("\twidget     :%.3f" %(time.time()-t));t=time.time()
        self.init_colorbars()              ;print("\tcb         :%.3f" %(time.time()-t));t=time.time()
        self.init_plots()           #;print("init_plots %.3f" %(time.time()-t));t=time.time()
        print(colors.green+'Completed'+colors.black)

    def load_data(self,df_name,Qdot,cols=['kdka','sigka','ka']):
        self.df_name=df_name
        self.Qdot=Qdot
        if Qdot:
            self.SphereArray = qsa.QdotSphereArray
        else:
            self.SphereArray = hsa.HardSphereArray
        self.df  = pd.read_pickle(df_name)       #; print(self.df)
        self.df['sigka'] = self.df.sigma/self.df.ka**2
        self.df['kdka'] = self.df.kd/self.df.ka
        self.cols0  = cols[:2]
        self.cb0col = cols[2]
        self.set_dats()
        d = self.df.loc[0]
        self.s = self.SphereArray(N=d.N,ka=d.ka,kp=d.kp,nmax=d.nmax-1,solve=True)
        self.s0 = copy.copy(self.s)
        self.s0.solve(copt=0)

    def set_dats(self):
        dats = self.df[self.cols0].values.T
        self.datsmM  = (dats[0].min(), dats[0].max(),  dats[1].min(), dats[1].max()) #;print(self.datsmM )
        self.dats    = self._normalize_dat(dats).T
        self.cb0vals = self.df[self.cb0col].unique()
    def _normalize_dat(self,dats):
        xm,xM,ym,yM = self.datsmM
        dats[0] = (dats[0] - xm)/(xM-xm)
        dats[1] = (dats[1] - ym)/(yM-ym)
        return dats

    def init_params(self):
        self.opts3 = 'tT'  #options for field to plot
        self.npts = 200
        self.r = (-1,1,1)   #range for extend of field to plot
        self.id0=10        #active case
        self.cmS,self.cmF,self.cmL = 'viridis', 'jet','Spectral'
        self.ip=0          #active sphere coeffs
        self.uncoupled = False
        self.fopts = [False]*3+[True,False]   #riam2
        self.cmax = 1

    def init_widgets(self):
        # interactive buttons,checkbox
        self.btns_cb=[self.open,self.set_ax0,self.change_settings,self.help]
        self.btns=[]
        for k,pos in btns_pos.items():
            self.btns += [self.fig.text(pos[0],pos[1],k,size=12,ha='left',va='top',backgroundcolor=(0.8,0.8,0.8))]
        self.plot_btn = self.fig.text(0.99,0.5,'click to plot nf'.ljust(20)+'!',size=12,ha='right',va='bottom',backgroundcolor='b',alpha=0.9)
        self.ckbtns = CheckButtons(self.fig.add_axes([0.955,0.5+hm,0.1,0.5-1.5*hm]), ['not\ncoupled','re','im',r'$\phi$','$||$','$||^2$'],[self.uncoupled]+self.fopts)
        self.ckbtns.on_clicked(self._ckbtns)

    def init_colorbars(self):
        # print('...init colorbars...')
        args = {'fig':self.fig,'is_3d':0,'cb':None,'cs':None}#,'shading':'auto'}
        cm, nmax = self.cmax, self.s.nmax
        nkps,kpM = self.cb0vals.size,self.cb0vals.max()

        self.cbs = np.zeros((4),dtype=object)
        self.cbs[0] = dsp.add_colorbar(ax=self.ax[0],cmap=self.cmS,caxis=[0,nkps],cb_pos=[0.455,0.5+hm,0.015,0.5-1.5*hm], values=np.arange(nkps), boundaries=np.arange(nkps)-0.5,ticks=np.arange(nkps),**args)
        self.cbs[1] = dsp.add_colorbar(ax=self.ax[1],cmap=self.cmL,caxis=[0,nmax],cb_pos=[0.455,hm    ,0.015,0.5-1.5*hm], values=np.arange(nmax), boundaries=np.arange(nmax)-0.5,ticks=np.arange(nmax),**args)
        self.cbs[3] = dsp.add_colorbar(ax=self.ax[3],cmap=self.cmF,caxis=[-cm,cm],cb_pos=[0.955,hm    ,0.015,0.5-1.5*hm], **args)
        self.cbs[1].ax.set_xlabel('$ l$'  ,color='k',fontsize=15)
        self.cbs[3].ax.set_xlabel('$ |f|$',color='k',fontsize=15)
        self.update_cb0()

        #interactive markers
        self.h0  = self.ax[0].plot(0,0,'s',color='k' ,markersize=12,animated=True)[0]
        self.h1  = self.ax[0].plot(0,0,'o',color='g' ,markersize=12,animated=True)[0]
        self.h11 = self.ax[1].plot(0,0,'o',color='g' ,markersize=5)[0]

    def update_cb0(self):
        nkps,kpM = self.cb0vals.size,self.cb0vals.max()
        tlabs = ['%.1f' %v for v in self.cb0vals]
        # tlabs = ['%g' %v for v in self.cb0vals]
        self.cbs[0].cmap = self.cmS
        self.cbs[0].max = nkps
        self.cbs[0].values = np.arange(nkps)
        self.cbs[0].boundaries = np.arange(nkps+1)-0.5
        self.cbs[0].ticks = np.arange(nkps)
        self.cbs[0].ax.set_xlabel('%s' %clab0[self.cb0col],color='k',fontsize=15)
        self.cbs[0].set_ticklabels(tlabs)
        self.cbs[0].update_ticks()

    # def update_cb3(self):
    def init_plots(self):
        self.clear()
        t=time.time()
        self.plot_s()                      ;print("\tplot_s     :%.3f" %(time.time()-t));t=time.time()
        self.plot_cp()                     ;print("\tplot_cp    :%.3f" %(time.time()-t));t=time.time()
        self.plot_fields(field=0)          ;print("\tplot_fields:%.3f" %(time.time()-t));t=time.time()
        self.update(button=1,hover=0,id0=0);print("\tupdate     :%.3f" %(time.time()-t));t=time.time()

    ##############################################################################################################
    ##### Update
    ##############################################################################################################
    def clear(self):
        for ax in self.ax:ax.cla()
        for i,ax in enumerate(self.ax):dsp.stddisp(ax=ax,**ax_props[i])
    def refresh(self):
        self.fig.canvas.draw()
        # self.bg = self.fig.canvas.copy_from_bbox(self.ax[0].bbox)
        self.bg = self.fig.canvas.copy_from_bbox(self.fig.bbox)
        print(colors.green+'figure refreshed'+colors.black)
    def update_ax0(self,i=0):
        self.fig.canvas.restore_region(self.bg)
        self.ax[i].draw_artist(self.h0)
        self.ax[i].draw_artist(self.h1)
        self.fig.canvas.update() #; self.fig.canvas.blit(self.fig.bbox)
        self.fig.canvas.flush_events()

    def update(self,button,hover,id0):
        d = self.df.loc[id0]
        col = self.csS[np.abs(self.cb0vals-d[self.cb0col]).argmin()]
        if button:
            if self.Qdot :
                cp = np.hstack([d.ap,d.bp])
            else:
                cp = d.Cp
            self.s = self.SphereArray(N=d.N,ka=d.ka,kp=d.kp,kd=d.kd,nmax=d.nmax-1,Cp=cp)
            if self.uncoupled:
                self.s0 = copy.copy(self.s)
                self.s0.solve(copt=0)
            self.id0 = id0                              #;t=time.time()
            self.ax[1].cla();self.plot_cp()             #;print('cp:',time.time()-t);t=time.time()
            self.ax[2].cla();self.plot_f()              #;print('ff:',time.time()-t);t=time.time()
            self.ax[3].cla();self.plot_fields(field=0)  #;print('nf:',time.time()-t);t=time.time()
            self.refresh()                              #;print('refresh:',time.time()-t);t=time.time()
            #update ax0
            self.h0.set_data(None,None)
            self.h1.set_data(d[self.cols0[0]],d[self.cols0[1]])#; print(self.h1)
            self.h1.set_color(col)                      #;print('misc:',time.time()-t);t=time.time()
            self.update_ax0()#self.h1)

        elif hover and not button:
            self.h0.set_data(d[self.cols0[0]],d[self.cols0[1]])
            self.h0.set_color(col)
            self.update_ax0()#self.h0)
        else:
            self.h0.set_data(None,None)
            self.update_ax0()#self.h0)



# import importlib as imp
import scipy.special as spe
import numpy as np
from . import displayStandards as dsp #;imp.reload(dsp)
from . import spherical_utils as spu #;imp.reload(spu)

class HardSphereArrayBase():
    def __init__(self,N=1,ka=1,kd=2,kp=np.inf,k=1,nmax=7,Cp=None,solve=False,copt=1):
        '''equally disitributed linear array of hard spheres
        - N,ka,kp,kd : nb spheres,wave number inside spheres, normalized radius, normalized distances
        - k : incident wave number
        - nmax : max included order expansion
        '''
        self.k,self.N,self.nmax = k,N,nmax+1
        self.ka,self.kp,self.kd,self.kd_p = ka,kp,kd,kd*np.arange(N)
        self.d_p  = self.kd_p/k
        if isinstance(Cp,np.ndarray):self.set_Cp(Cp)
        if solve:self.solve(copt=copt)

    def get_s(self,npts=361):
        theta = np.linspace(0,np.pi,npts)
        dt = theta[1]-theta[0]
        f = self.get_ff(theta)
        s = np.sum(np.abs(f)**2*np.sin(theta))*2*np.pi*dt
        return s

    ###############################################################
    #  display
    ###############################################################
    def show_f(self,cart_opt=True,npts=200,r=None,opts='tT',idp=None,fz=np.abs,name='',def_args=True,**kwargs):
        '''
        opts : i(incident), s(scattered), t(total), P(sphere idp),T(All), G(Gradient)
        '''
        ka,kd,N = self.ka,self.kd,self.N
        if not isinstance(r,tuple) or not isinstance(r,list) or not isinstance(r,np.ndarray):
            r=(1e-3,4*ka+N*kd,-2*ka,N*kd+2*ka)
        r,theta,y,z = spu.polar_mesh2D(cart_opt,npts,r)
        k,N,nmax = self.k,self.N,self.nmax
        ap,dp = self.ka/k,self.kd/k

        if idp:self._check_idp(idp=None)
        args = {}
        if def_args:
            args = {'lw':2,'labs':['$y$','$z$'],'imOpt':'c','axPos':'V','fonts':{'title':25},}


        t = np.linspace(-np.pi/2,np.pi/2,100)
        ct,st = np.cos(t), np.sin(t)
        plts = [ [ap*ct, dp*p+ap*st,'k-',''] for p in range(N)]
        fstr =  r'$%s \psi(r,\theta)$' %(['',r'\partial_r']['G' in opts])
        name+=['f','df']['G' in opts]
        if 'P' in opts:
            for p in idp:
                print("...compute amplitude at p=%d..." %p)
                fi,fs = self.compute_f(r,theta,0,ftype='a',Gopt='G' in opts,idp=p)
                if 'i' in opts:
                    f = fz(fi)
                    dsp.stddisp(plts,im=[y,z,f],
                        title = r"Incident %s at sphere %d, $N_{max}=%d$" %(fstr,p,nmax),
                        name=name+'i_sphere%d.png' %p,**args,**kwargs)
                if 's' in opts:
                    f = fz(fs)
                    dsp.stddisp(plts,im=[y,z,f],
                        title = r"Scattered %s at sphere %d, $N_{max}=%d$" %(fstr,p,nmax),
                        name=name+'s_sphere%d.png' %p,**args,**kwargs)
                if 't' in opts:
                    f = fz(fi+fs)
                    dsp.stddisp(plts,im=[y,z,f],
                        title = r"Total %s at sphere %d, $N_{max}=%d$" %(fstr,p,nmax),
                        name=name+'t_sphere%d.png' %p,**args,**kwargs)
        if 'T' in opts:
            print("...compute total amplitude ...")
            fi,fs = self.compute_f(r,theta,0,ftype='a',Gopt='G' in opts)
            params = self._params()
            if 'i' in opts:
                f = fz(fi)
                dsp.stddisp(plts,im=[y,z,f],
                    title = r"Incident %s, %s " %(fstr,params),
                    name=name+'i.png',**args,**kwargs)
            if 's' in opts:
                f = fz(fs)
                dsp.stddisp(plts,im=[y,z,f],
                    title = r"Scattered %s, %s" %(fstr,params),
                    name=name+'s.png',**args,**kwargs)
            if 't' in opts:
                f = fz(fi+fs)
                dsp.stddisp(plts,im=[y,z,f],
                    title = r"Total %s, %s" %(fstr,params),
                    name=name+'t.png',**args,**kwargs)
        return f.max(),f.min()

    def show_ff(self,npts=361,fopts='m',lw=2,leg=1,name='',title=None,**kwargs):
        ''' display far field scattering amplitude
        - fopts : r(real), i(imag), m(mag), a(angle), 2(mag2)
        '''
        theta  = np.linspace(0,np.pi,npts)
        theta_d = np.rad2deg(theta)
        ff = self.get_ff(theta)


        plts,ls = [], ['--','-'][leg]
        if 'r' in fopts: plts+= [[theta_d,np.real(ff)    ,['c'          ,ls],['',r'$\Re$' ][leg] ]]
        if 'i' in fopts: plts+= [[theta_d,np.imag(ff)    ,['r'          ,ls],['',r'$\Im$' ][leg] ]]
        if 'a' in fopts: plts+= [[theta_d,np.angle(ff)   ,['m'          ,ls],['',r'$\phi$'][leg] ]]
        if 'm' in fopts: plts+= [[theta_d,np.abs(ff)     ,['b'          ,ls],['',r'$||$'  ][leg] ]]
        if '2' in fopts: plts+= [[theta_d,np.abs(ff)**2  ,[[0.25,0.75,1],ls],['',r'$||^2$'][leg] ]]
        if not isinstance(title,str):
            title='Scattering amplitude for %s' %self._params()
        return dsp.stddisp(plts,labs=[r"$\theta(^{\circ})$",r"$f(\theta)$"],lw=lw,
            title=title, xylims=[0,180,0,np.abs(ff).max()],
            name=name+'fka.svg',**kwargs)

    def _params(self):
        ss = '$N=%d$, $n_{ref}=%g$, $ka=%.1f$, $kd=%.1f$' %(self.N,self.kp, self.ka,self.kd)
        return ss
    ############################################################################
    # misc
    ############################################################################
    def _check_idp(self,idp=None):
        if isinstance(idp,int) :
            if idp>=0 and idp<self.N:
                idp=[idp]
            else:
                print(colors.red+'warn:sphere index not within 0<=idp<N'+colors.black)
                idp=np.arange(self.N)
        else:
            idp=np.arange(self.N)
        return idp

    def test_convergence(self,nmaxs,npts=361,name='', **kwargs):
        ns = np.array(nmaxs).size
        theta = np.linspace(0,np.pi,npts)
        f = np.zeros((ns,npts),dtype=complex)
        for i,nmax in enumerate(nmaxs):
            print('solving nmax=%d' %nmax)
            self.solve(nmax=nmax,copt=1,v=0)
            f[i] = self.get_ff(theta)
        theta_d = np.rad2deg(theta)

        #### display
        cs   = dsp.getCs('Spectral',ns)
        nref = ['%g' %self.kp,r'\infty'][self.kp==np.inf]
        tle  = 'Convergence test $ka=%.1f$, $kd=%.1f$, $n_{ref}=%s$' %(self.ka,self.kd,nref)
        labs =[r"$\theta(^{\circ})$",r"$|f(\theta)|$"]
        plts = [[theta_d,np.abs(f[i]), cs[i] ,'$n_{max}=%d$' %nmax] for i,nmax in enumerate(nmaxs)]
        dsp.stddisp(plts,labs=labs,lw=2,
            title=tle,xylims=['x',0,180],
            name=name+'fconv.svg',**kwargs)



# import importlib as imp
import scipy.special as spe
import numpy as np, pandas as pd
from . import spherical_utils as spu  #;imp.reload(spu)
from . import hard_sphere_base as hsb #;imp.reload(hsb)
from . import glob_colors as colors

Pl = lambda l,theta: np.sqrt(4*np.pi/(2*l+1))*spe.sph_harm(0,l,0,theta)

class HardSphereArray(hsb.HardSphereArrayBase):
    def solve(self,nmax=-1,copt=1,v=1):
        if nmax>0:self.nmax=nmax+1
        N,ka,kd,kdp,nmax = self.N,self.ka,self.kd,self.kd_p,self.nmax
        fz_q = spu.hn1

        #preprocess
        ll = np.arange(nmax)
        bl = 1J**ll*(2*ll+1)*np.sqrt(4*np.pi/(2*ll+1))
        ul = -np.array([spu.jn(l,ka)/spu.hn1(l,ka) for l in range(nmax)])

        #### incident plane wave response
        L = np.tile(bl*ul,N)
        for p in range(N):
            idpl = p*nmax+np.arange(nmax) #;print('p=%d'%p)#;print('idp:' , idpl)
            L[idpl] *= np.exp(1J*kdp[p])

        #### coupling
        T = np.zeros((N*nmax,N*nmax),dtype=complex)
        if copt:
            ul = ul[:,None]
            if v:print("...assembling coupling...")
            for p in range(N):
                idpl = p*nmax+np.arange(nmax) #;print('p=%d'%p)#;print('idp:' , idpl)
                for q in range(p):
                    kdpq = kdp[p]-kdp[q]
                    aln0 = a_ln(nmax-1,nmax-1, fz_q,kdpq,np.pi)
                    T[p*nmax:(p+1)*nmax,q*nmax:(q+1)*nmax] = ul*aln0
                for q in range(p+1,N):
                    kdpq = kdp[q]-kdp[p]
                    aln1 = a_ln(nmax-1,nmax-1, fz_q,kdpq,0*np.pi)
                    T[p*nmax:(p+1)*nmax,q*nmax:(q+1)*nmax] = ul*aln1

        if v:print("...solving...")
        self.Cp = np.linalg.solve(np.identity(N*nmax)-copt*T,L)
        return self.Cp

    def set_Cp(self,Cp):self.Cp=Cp

    def compute_f(self,r,theta,phi,ftype='t',Gopt=0,idp=None):
        ''' computes scattering amplitude f
        - r,theta,phi : np.ndarray each - coordinates
        - ftype : str - 't'(total), 's'(scattered), 'i'(incident)
        - idp : index of sphere to show (None or -1 =>all)
        return :
        - np.real(E_ftype)
        '''
        k,d_p,nmax,N = self.k,self.d_p,self.nmax,self.N
        x,y,z = spu.sphere2cart(r,theta,phi)
        idp = self._check_idp(idp) #;print(idp)

        #incident wave
        if ftype in 'ita':
            fi = np.zeros(r.shape,dtype=complex)
            # plane wave at sphere idp : use local spherical decomposition
            if isinstance(idp,list):
                # print('incident field at p sphere')
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[idp[0]])
                for l in range(nmax):
                    fi += 1J**l*(2*l+1)*spu.jn(l,k*r_p)*Pl(l,theta_p)
                fi *= np.exp(1J*k*self.d_p[idp])
            # Otherwise actual plane wave propagating along z
            else:
                # print('full incident field')
                fi = np.exp(1J*k*z)

        #scattered fields
        if ftype in 'sta':
            fs = np.zeros(r.shape,dtype=complex)
            for p in idp:
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
                for l in range(nmax):
                    fs += self.Cp[p*nmax+l]*spu.hn1(l,k*r_p)*spe.sph_harm(0,l,0,theta_p)

        #remove values inside pth spheres
        for p in range(N):
            r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
            idx_p = r_p<self.ka/k
            if ftype in 'sta' : fs[idx_p] = 0
            if ftype in 'ita' : fi[idx_p] = 0

        #total field
        if   ftype=='t' :return fs+fi
        elif ftype=='s' :return fs
        elif ftype=='i' :return fi
        elif ftype=='a' :return fi,fs

    def get_ff(self,theta):
        k,d_p,nmax,N = self.k,self.d_p,self.nmax,self.N
        ct = np.cos(theta)

        f = np.zeros(theta.shape,dtype=complex)
        for l in range(nmax):
            Yl = spe.sph_harm(0,l,0,theta)
            for p in np.arange(N):
                f += (-1J)**(l+1)*Yl*self.Cp[p*nmax+l]*np.exp(-1J*self.kd_p[p]*ct)
        return f

def a_ln(lmax,nmax, fz_q,r_d,theta_d):
    '''Scalar translational addition theorem coefficients
    - Nmax : order of expansion
    - fz_q : type of expansion (bessel or hankel)
    - r_d,theta_d : translation vector
    '''
    l = np.arange(lmax+nmax+2)
    Yq  = np.sqrt((2*l+1)/(4*np.pi))*spe.lpn(lmax+nmax+1,np.cos(theta_d))[0]
    # Yq = np.array([spe.sph_harm(0,q,0,theta_d) for q in range(lmax+nmax)])
    zq = np.array([fz_q(q,r_d) for q in range(lmax+nmax+2)])

    aln = np.zeros((lmax+1,nmax+1),dtype=complex)
    for l in range(lmax+1):
        # print('l=%d' %l)
        for n in range(nmax+1):
            q = np.arange(abs(l-n),n+l+1)   #;cput = time.time()
            Glnq = np.array([np.float(spu.gaunt(l,n,q, 0,0,0)) for q in q])
            # Glnq = spu.w3j(l,n,q, 0,0,0)
            aln[l,n] = 4*np.pi*np.sum((-1J)**(l-n-q)*zq[q]*Yq[q]*Glnq) #;print('n=%d' %n, q,zq[q],Yq[q],Glnq)
    return aln


def sweep_ka(df_name,N=2,nmax=7,kas=[0.5,1,2,4],kds=None,nkds=50,kdr=(2,10),v=1):
    ''' solve for a series of values kas,kds and save the Cp
    - df_name : name of data to create or update
    - kas : array of ka to solve
    - kds : array of kd to solve
    - kdr,nkds : kds=np.linspace(kdr[0]*ka,kdr[1]*ka,nkds) (only if kds=None)
    '''
    kas = np.array(kas)
    nkas = kas.size
    kds_var=not (isinstance(kds,np.ndarray) or isinstance(kds,list))
    cols = ['N','ka','kd','nmax','Cp','sigma']
    df = pd.DataFrame([],columns=cols)
    for i,ka in enumerate(kas):
        if v:print('...%d : solving ka=%.1f...' %(i,ka))
        if kds_var: kds = np.linspace(kdr[0]*ka,kdr[1]*ka,nkds)
        for j,kd in enumerate(kds):
            if kd>2*ka:
                s = HardSphereArray(N,ka,kd,nmax,solve=True) #s;print(s.Cp)
                di = dict(zip(cols,[s.N,s.ka,s.kd,s.nmax,s.Cp,s.get_s()])) #;print(di)
                df = df.append(di,ignore_index=True)

    try :
        df0 = pd.read_pickle(df_name)
        df.update(df0)
    except FileNotFoundError:
        print('creating new file')
    df.to_pickle(df_name)
    print(colors.green+'DataFrame saved : \n'+colors.yellow+df_name+colors.black)
    return df



'''Display utilities'''
import numpy as np
import matplotlib,os
import matplotlib.pyplot as plt
# import matplotlib.animation as animation
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.lines import Line2D
from matplotlib.collections import PatchCollection as PatchColl
from matplotlib.patches import Rectangle
from scipy import ndimage
from subprocess import check_output
from PIL import Image
# from .glob_colors import*
from . import glob_colors as colors
# Get screen info, remove toolbar
# matplotlib.rcParams['toolbar'] = 'None'
matplotlib.rcParams['backend']='GTK3Agg'
# matplotlib.rcParams['pcolor.shading'] = 'auto'
# matplotlib.rc('backend','GTK3Agg')
matplotlib.rc('text', usetex=True)
# matplotlib.rcParams['text.latex.preamble']=[r"\usepackage{amsmath}"]
dpi = check_output("xdpyinfo | awk '/resolution/{print $2}'",shell=True).decode()
dpi = np.array(dpi.strip().split('x'),dtype=int)
screen_size = check_output("xrandr | grep \"*+\" | awk '{print $1}'",shell=True).decode().split('\n')
#choose second monitor if present
screen_size=screen_size[int(len(screen_size)>2)]
screen_size = np.array(screen_size.split('x'),dtype=int)/dpi #inches


###########################################################################
#def : plotting standard
###########################################################################
def standardDisplay(ax,labs=['','',''],name='', xylims=[], axPos=1,legOpt=None,view=None,
    fonts=[30,25,15,20], c=['k','k'], logOpt='',changeXYlims=True,is_3d=0,
    gridOn=True,gridmOn=False,ticksOn=True,title='',legLoc='best',legElt=[],
    figopt='1',xyTicks=None,xyTicksm=None,xyTickLabs=None,mg=0.05,opt='p',setPos=True,equal=False,
    pOpt=None):
    '''
    legElt : dict - {'lab1':[(1,0,0),'s-']} overrides existing handles
    opt : p(plot), s(save), c(close) '
    pOpt : p(setPos)X(changeXYlims)l(legOpt)t(ticksOn)e(equal),G(gridOn),g(gridmOn)
    view : [elev,azim] for 3D projection axes
    '''
    is_3d = "3D" in str(ax.__class__)
    if isinstance(pOpt,str):
        setPos,changeXYlims,legOpt,ticksOn,equal,gridOn,gridmOn = [s in pOpt for s in 'pXlteGg']
    axPos = get_axPos(axPos); #print(changeXYlims)
    ax.name = name;
    if isinstance(fonts,dict) : fonts = get_font(fonts,dOpt=True)
    fs,fsLeg,fsL,fsT = fonts; #print(fsLeg)
    #axis, labels, position
    if 'x' in logOpt : ax.set_xscale('log')
    if 'y' in logOpt : ax.set_yscale('log')
    ax.set_xlabel(labs[0],fontsize=fs   ,color=c[0])
    ax.set_ylabel(labs[1],fontsize=fs   ,color=c[1])
    if is_3d : ax.set_zlabel(labs[2],fontsize=fs   ,color=c[1])
    ax.axis(['off','on'][ticksOn])
    ax.set_axisbelow(True)
    if setPos : ax.set_position(axPos);
    #lims,ticks and grid
    xylims=changeAxesLim(ax,mg,xylims,is_3d,changeXYlims)#;print(changeXYlims,xylims)
    change_ticks(ax,xyTicks,xyTickLabs,xylims,is_3d,xyTicksm)
    ax.tick_params('x',labelsize=fsL    ,colors=c[0],direction='in')
    ax.tick_params('y',labelsize=fsL    ,colors=c[1],direction='in')
    if is_3d : ax.tick_params('z',labelsize=fsL,colors=c[1])
    ax.grid(False);#print(gridOn,gridmOn)
    if gridOn  : ax.grid(gridOn,which='major',color=(0.9,0.9,0.9),linestyle='-')
    if gridmOn :
        ax.minorticks_on()
        ax.grid(gridmOn,which='minor',color=(0.95,0.95,0.95),linestyle='--')
    if view : ax.view_init(elev=view[0], azim=view[1])
    #title,legend,save
    if equal :ax.axis(['equal','auto'][is_3d]);
    ax.set_title(title, {'fontsize': fsT}); #print(title)
    addLegend(ax,fsLeg,legOpt,legLoc,legElt)
    disp_quick(name,ax,opt,figopt)

def stddisp(plots=[],scat=None,texts=[],colls=[],patches=[],im=None,rot=0,
            tricont=None,contour=None,quiv=None,surfs=[],caxis=None,
            cs=None,lw=1,ms=5,marker='o',fonts={},axPos=1,imOpt='',cb_pos=[],cmap='jet',
            ax=None,fig=None,cb=None,figsize=(0.5,1),pad=None,rc='11',
            inset=None,iOpt='Gt',name='',opt='p',figopt='',pargs={},
            std=1,**kwargs):
    '''
    fonts.keys=['lab','leg','tick','text','title']
    lw,ms,marker : linewidth,markersize,marker to apply to all
    axPos   : 4-list or int - ax position (see get_axPos)
    contour : list - x,y,z,[levels]
    quiv    : functor to gradient that will be plotted on the isocontours (must return shape (2,N,N))
    imOpt   : c(colorbar) h(horizontal) v(vertical)
    cs      : C(contours),I(image),S(scatter),N(None) artist to use for colorbar
    iOpt    : inset displayStandards keys - ['axpos','xylims','ms','lw','lwp']
    kwargs  : standardDisplay
    '''
    if isinstance(fonts,dict) : fonts = get_font(fonts)
    fsT = fonts[3]; fonts=(np.array(fonts)[[0,1,2,4]]).tolist()
    if not ax : fig,ax = create_fig(figsize,pad,rc)
    if ax and not axPos : axPos=ax.get_position().bounds
    #add lines,patches,texts,scatters
    is_3d = "3D" in str(ax.__class__) # not isinstance(ax,matplotlib.axes._subplots.Subplot) and not isinstance(ax,matplotlib.axes._axes.Axes)#;print(is_3d)
    for coll in colls : ax.add_collection(coll)
    for pp in patches : ax.add_patch(pp)
    pltPlots(ax,plots,lw,ms,is_3d,pargs)
    pltTexts(ax,texts,fsT,is_3d)

    #note that at them moment only one cb per ax authorized
    # priority order here : scat,im,contour
    css = dict(zip(['I','S','C','N'],[None]*4))
    if not contour==None:css['C'] = plt_contours(ax,contour,quiv,cmap,lw)
    if not im==None:css['I']      = pltImages(ax,im,cmap,caxis,imOpt,rot)
    if not scat==None:css['S']    = pltScatter(ax,scat,'b',ms,marker,rc=='3d',cmap)
    # if 'c' in imOpt and contour==None and scat==None and im==None:cs=None
    if isinstance(cs,str) : cs=css[cs]
    add_colorbar(fig,ax,cs,cb,imOpt,cmap,is_3d,caxis=caxis,cb_pos=cb_pos)
    if is_3d :pltSurfs(ax,surfs)

    if isinstance(inset,dict):
        inset=fill_inset_dict(inset)
        ec = inset['ec']
        in_box = inset['xylims']
        xy,Wrect,Hrect = (in_box[0],in_box[2]),in_box[1]-in_box[0],in_box[3]-in_box[2]
        ax.add_patch(Rectangle(xy,Wrect,Hrect,fill=0,ec=ec,lw=inset['lwp']))
        ax2 = add_inset(fig,plots,inset['xylims'],inset['axpos'],inset['lw'],inset['ms'],iOpt)
        for axis in ['top','bottom','left','right']:
            ax2.spines[axis].set_linewidth(3)
            ax2.spines[axis].set_color(ec)
    if std:
        standardDisplay(ax,is_3d=is_3d,axPos=axPos,fonts=fonts,
            name=name,opt=opt,figopt=figopt,**kwargs)
    else:
        disp_quick(name,ax,opt,figopt)
    return fig,ax


def image_bg(im,rot=0,name='',opt='p',figopt='',**kwargs):
    ''' display an image in background to data
    fig = dsp.image_bg('image.png',xylims=[0,1,0,1])
    NOTE :
    use fix_pos(fig) after the figure has been displayed
    '''
    fig,ax = stddisp(im=im,pOpt='',opt='',rot=rot)
    ax1 = fig.add_axes([0,0,1,1],frameon=False)
    fig,ax1 = stddisp(fig=fig,ax=ax1,pOpt='GtX',opt='',**kwargs)
    fix_pos(fig)
    disp_quick(name,ax,opt,figopt)
    return fig

def fix_pos(fig):
    fig.axes[1].set_position(fig.axes[0].get_position())
    fig.canvas.draw()
    return fig

def display_solutions(disp_d,objs=[],key_pair=(),help=0,**kwargs):
    ''' display a pair data from different objects (used to compare methods)
    - disp_d : dict - {key_sol : (sol_function_str,color,display_label))
    - objs : list of tuples - [(obj,cml,leg)...] or [obj...]
    - key_pair : (str,str or list of str) where str in disp_d.keys()
    - help : for the list of available keys
    '''

    if help :print('available keys:\n',disp_d.keys());return


    kx,kys = key_pair[0],key_pair[1]
    fx_str,c,xlab = disp_d[kx]
    if isinstance(kys,str):kys=[kys] #ylab = disp_d[al[kys]][0]

    #legend for the quantities displayed
    legElt = {disp_d[ky][2]:disp_d[ky][1] for ky in kys}

    # keys legend
    plts,ylab = [],''
    for obj_t in objs:
        if isinstance(obj_t,tuple):
            obj,cml,leg = obj_t
            legElt[leg] = cml
            ml = ''.join(getCML(cml)[1:3])
        else:
            obj,ml=obj_t,'-'
        for ky in kys:
            fy_str,c = disp_d[ky][0:2]
            fx = obj.__getattribute__(fx_str)
            fy = obj.__getattribute__(fy_str)
            plts += [[fx(),fy(),[c,ml],'']]

    #display
    labs = [xlab,ylab]
    fig,ax = stddisp(plts,labs=labs,legElt=legElt,**kwargs)
    return fig,ax,labs,legElt

def add_colorbar(fig,ax,cs=None,cb=None,imOpt='c',cmap='jet',is_3d=0,l=0.03,L=0.85,caxis=None,cb_pos=[],**kwargs):
    if 'c' in imOpt :
        if not cs :
            if caxis==None:caxis=[0,1]
            cs = plt.cm.ScalarMappable(plt.Normalize(vmin=caxis[0],vmax=caxis[1]),cmap=cmap)
        if is_3d:
            cb = fig.colorbar(cs,ax=ax,vmin=caxis[0],vmax=caxis[1],cmap=cmap)
        else:
            if cb:
                cb.mappable=cs
            else:
                if not cb_pos : cb_pos = [[0.9,0.1,l,L],[0.1,0.9,L,l]]['h' in imOpt]
                orient,tickloc=[['vertical','right'],['horizontal','top']]['h' in imOpt]
                ax_cb = fig.add_axes(cb_pos)
                cb=fig.colorbar(cs,ax=ax,cax=ax_cb,orientation=orient,ticklocation=tickloc,**kwargs)
        return cb
#######################################################################
###### Old one
def stdDispPlt(plots,xlabs=['',''],name='',xlims=[],axPos=[],c=['k','k'],
               showOpt=1,ax=None,fig=None,texts=[],legOpt=1,lw=1,ms=5,fonts={},
               logOpt='',changeXYlims=True,gridOn=True,ticksOn=True,fullsize=False,title='',
               legLoc='upper left',legElt=[],xyTicks=[],colls=[],opt='',figsize=(9,9)):
    '''fonts.keys=['lab','leg','tick','text','title']'''
    if isinstance(fonts,dict) : fonts = get_font(fonts)
    fsT = fonts[3]; fonts=(np.array(fonts)[[0,1,2,4]]).tolist()
    if showOpt and not 'q' in opt : opt+='p'
    if not ax : fig,ax = plt.subplots(figsize=figsize)
    if ax and axPos==[] : axPos=ax.get_position().bounds
    if fullsize : mng = plt.get_current_fig_manager(); mng.resize(*mng.window.maxsize())
    for coll in colls : ax.add_collection(coll)
    pltPlots(ax,plots,lw,ms)
    pltTexts(ax,texts,fsT)
    standardDisplay(ax,xlabs,name,xlims,axPos ,legOpt,fonts,c,
                    logOpt=logOpt,changeXYlims=changeXYlims,gridOn=gridOn,title=title,
                    legLoc=legLoc,legElt=legElt,xyTicks=xyTicks,opt=opt)
    return fig,ax

def addyAxis(fig,ax,plots,yLab='',c='k', lw=1,ms=5,axPos=[],showOpt=1,yTicks=[],yTickLabs=[],
             **kwargs):
    axPosI = axPos if len(axPos)==4  else [0.13, 0.15, 0.75, 0.75]
    ax2 = ax.twinx()
    pltPlots(ax2,plots,lw)
    standardDisplay(ax2,['',yLab],c=['k',c],axPos=axPosI,xyTicks=[[],yTicks],xyTickLabs=[[],yTickLabs],
                    **kwargs)
    ax.set_position(axPosI)
    # if showOpt : fig.show()

def addxAxis(ax,plots,xLab='',c='k', lw=1,axPos=[],
            xTicks=[],xTickLabs=[], **kwargs):
    axPosI = axPos if len(axPos)==4  else [0.13, 0.15, 0.8, 0.675]
    ax2 = ax.twiny()
    pltPlots(ax2,plots,lw)
    standardDisplay(ax2,[xLab,''],c=[c,'k'],axPos=axPosI,xyTicks=[xTicks,[]],xyTickLabs=[xTickLabs,[]],**kwargs)
    ax2.tick_params('x',color=c)
    ax2.set_xlabel(xLab, verticalalignment='bottom')
    ax.set_position(axPosI)
    #if showOpt : plt.show()

def add_inset(fig,plots,xylims,axPosI,lw=2,ms=5,iOpt='GX'):#,**kwargs)
    ax2 = fig.add_axes(axPosI)
    pltPlots(ax2,plots,lw,ms) #; print(xylims)
    standardDisplay(ax2,labs=['',''],xylims=xylims,setPos=False,
        xyTicks=None,xyTickLabs=[[],[]],legOpt=0,pOpt=iOpt)#,**kwargs)
    return ax2

########################################################################
### def : plot calls
########################################################################
def pltPlots(ax,plots,lw0,ms0=5,is_3d=0,pargs={}):
    ''' plots a list of plots :
        - [x,y,color]([x,y,z,color] if is_3d) + [label](optional)
    '''
    if plots :
        if len(plots)>=3+is_3d and isinstance(plots[2+is_3d],str):
            plots = [plots]; #print(len(plots))
    if is_3d :
        for p in plots:
            cml = 'b' if len(p)<4 else p[3]
            lab = ''  if len(p)<5 else p[4]
            c,m,l = getCML(cml)
            ax.plot(p[0],p[1],p[2],label=lab,color=c,linestyle=l,marker=m, linewidth=lw0,markersize=ms0)
    else:
        for p in plots:
            cml = 'b' if len(p)<3 else p[2]
            lab = ''  if len(p)<4 else p[3]
            lw  = lw0 if len(p)<5 else p[4]
            c,m,l = getCML(cml)
            ax.plot(p[0],p[1],label=lab,color=c,linestyle=l,marker=m, linewidth=lw,markersize=ms0,**pargs)
def pltTexts(ax,texts,fsT,is_3d=0):
    ''''texts format : [x,y,text,color], [x,y,z,txt,color] if is_3d'''
    if texts:
         if not isinstance(texts[0],list) : texts = [texts]
    if is_3d :
        for t in texts:
            ax.text(t[0],t[1],t[2],t[3],fontsize=fsT,color=t[4])
    else:
        for t in texts:
            c_t = 'k' if len(t)<4 else t[3]
            ax.text(t[0],t[1],t[2],fontsize=fsT,color=c_t)
def pltImages(ax,im=None,cmap='viridis',caxis=None,imOpt='',rot=0):
    cs = None
    if isinstance(im,str):
        if '.npy' in im:
            im = np.load(im)
            if len(im.shape)==3:im=list(im)
        else:
            if 'g' in imOpt:
                image = Image.open(fname).convert("L")
                image = np.asarray(image)/255
                cs = ax.pcolor(im,cmap='gray',)
                im = None
            else:
                image = plt.imread(im)
                if rot:image=ndimage.rotate(image,rot)
                cs=ax.imshow(image,cmap=cmap)#,**im_args)#,origin='upper')
                im = None
    if isinstance(im,np.ndarray): im=[im]
    if isinstance(im,list):
        if len(im)>2:
            x,y,z = im[:3]
            N = int(np.sqrt(x.shape[0]))
            # dx,dy = abs(x[0,0]-x[0,1])/2, abs(y[0,0]-y[1,0])/2
            args = {}
            if caxis : args = {'vmin':caxis[0],'vmax':caxis[1]}
            if len(im)>3:args['alpha']=im[3]
            # cs=ax.pcolormesh(x-dx,y-dy,z,cmap=cmap,edgecolors='none',ec=None,**args)
            cs=ax.pcolormesh(x,y,z,shading='auto',cmap=cmap,edgecolors='none',ec=None,**args)
        else :
            # cs=ax.pcolor(im[0],cmap=cmap)
            cs=ax.imshow(im[0],cmap=cmap)
    return cs
def pltScatter(ax,scat,c='b',s=5,marker='o',proj_3D=False,cmap='jet') :
    '''
    - scat : [x,y,<z>,s,c] or [x,y,<z>,c]
    - s : int or list/np.array of int
    - c : tuple,str or list/np.array of tuple/str
    '''
    cs = None
    if len(scat) :
        if proj_3D:
            x,y,z = scat[:3]
        else :
            x,y = scat[:2]
        # color and marker size
        sc = scat[2+proj_3D:]
        if len(sc)==1 : c = sc[0]
        elif len(sc)==2: s,c = sc
        N=np.array(x).size
        if isinstance(s,int) : s = [s]*N
        if isinstance(c,tuple) or isinstance(c,str) : c = [c]*N
        if proj_3D :
            cs=ax.scatter3D(x,y,z,s=s,c=c,marker=marker,cmap=cmap)
        else :
            cs=ax.scatter(x,y,s,c,marker=marker,cmap=cmap)
    return cs

def pltSurfs(ax,surfs,c='b',a=0.2,lw=1,ec='b'):
    '''surf : [x,y,z] or [x,y,z,**kwargs] or [x,y,z,c,a,lw,ec]
    '''
    for surf in surfs:
        x,y,z = surf[:3]
        if len(surf)>3 :
            if isinstance(surf[3],dict):
                kwargs=surf[3]
                ax.plot_surface(x,y,z,**kwargs)
                return
            else:
                c,a,lw,ec = surf[3:]
        ax.plot_surface(x,y,z,color=c,alpha=a,linewidth=lw,edgecolor=c)

def plt_contours(ax,contour,quiv,cmap,lw):
    cs = ax.contour(*contour,cmap=cmap,linewidths=lw)
    if not quiv == None:
        v_cs = get_iso_contours_vertx(cs)
        x_cs,y_cs = v_cs.T
        dfx,dfy = quiv(x_cs,y_cs)
        ax.quiver(x_cs,y_cs,dfx,dfy,linewidths=lw,units='xy')
    return cs

########################################################################
# handles and properties
########################################################################
def create_fig(figsize=(0.5,1),pad=None,rc='11') :
    '''figsize :
        tuple : (width,height) normalized units
        str   : f(full),12(half),22(quarter)
        rc    : str or list - layout arrangement : '3d','11','22',...
    '''
    if isinstance(figsize,str) :
        figsize = {'f':(1,1),'12':(0.5,1),'21':(1,0.5),'22':(0.5,0.5)}[figsize]
    if isinstance(rc,str) :
        rc = {'fig':'fig','3d':'3d','11':[1,1],'21':[2,1],'12':[1,2],'22':[2,2]}[rc];#print(rc)
    figsize = tuple(np.array(figsize)*screen_size)
    if rc=='3d':
        wh = min(screen_size)
        figsize = (wh,wh);#print(figsize)
        fig = plt.figure(figsize=figsize,dpi=dpi[0])
        ax  = plt.subplot(111,projection='3d')
    elif rc=='fig':
        fig=plt.figure(figsize=figsize,dpi=dpi[0])
        return fig
    else:
        fig,ax = plt.subplots(nrows=rc[0],ncols=rc[1],figsize=figsize,dpi=dpi[0])
    ax.set_navigate(True)
    if pad : plt.tight_layout(pad)
    return fig,ax

def get_font(d_font=dict(),dOpt=False) :
    '''keys = ['lab','leg','tick','text','title']
    '''
    keys = ['lab','leg','tick','text','title']
    # vals = [30,25,15,20,30]
    vals = [25,20,15,20,20]
    font_dict = dict(zip(keys,vals))
    for k in d_font.keys() : font_dict[k] = d_font[k]
    if dOpt : keys = ['lab','leg','tick','title']
    fonts = [ font_dict[k] for k in keys]
    return  fonts

def fill_inset_dict(inset_dict=dict()):
    '''keys = ['axpos','xylims','ms','lw','lwp']'''
    keys = ['axpos','xylims','ms','lw','lwp','ec']
    vals = [[0,0,0.25,0.25],None,30,3,2,'g']
    full_dict = dict(zip(keys,vals))
    for k,v in inset_dict.items() : full_dict[k]=v
    return full_dict

def getCML(C):
    ml = ''
    if isinstance(C,list):
        c,ml = C[0], C[1]
    elif isinstance(C,tuple) or isinstance(C,np.ndarray):
        c,m,l = C, '','-'
    elif isinstance(C,str):
        c,m,l,ml = C[0],'','-',C[1:]
    else:
        c,m,l= 'b','','-'

    if not ml == '':
        l = [char for char in ml if char in '-.:']    ;l=''.join(l)
        m = [char for char in ml if char not in '-.:'];m=''.join(m)
    return c,m,l

def get_legElt(legElt):
    legE=[]
    if isinstance(legElt,dict):
        for lab,cml in legElt.items():
            c,m,l=getCML(cml)
            legE+=[Line2D([0],[0],linewidth=2,linestyle=l,color=c,marker=m,label=lab)]
    return legE

def addLegend(ax,fsLeg,legOpt=None,loc='upper left',legElt=[]):
    hs,labs = ax.get_legend_handles_labels()
    legElt = get_legElt(legElt)
    if legOpt == None : legOpt = any(labs) or any(legElt)
    if legOpt:
        out = None
        if loc[-4:]==' out' :
            loc = loc[:-4]
            if loc=='upper left' : out = (1, 1)
            elif loc=='center left' : out = (1, 0.5)

        leg = ax.legend(handles=hs+legElt,fontsize=fsLeg,loc=loc,bbox_to_anchor=out)
        leg.set_draggable(True)         # matplotlib 3.1
        #leg.draggable(True)            # matplotlib 2.2
        #if leg : leg.DraggableLegend() # matplotlib?

def getCs(name,N,cmopt=0):
    if cmopt:
        cs = matplotlib.cm.get_cmap(name)
    else:
        cmap = matplotlib.cm.get_cmap(name)
        cs = [cmap(float(i+1)/N)[0:3] for i in range(N)]
    return cs

def get_iso_contours(CS):
    '''get the list of iso-contours coordinates '''
    coords = [c.get_paths()[0].vertices for c in CS.collections]
    return coords

def get_iso_contours_vertx(cs):
    '''Get all vertices from isocontours
    returns Nx2 ndarray
    '''
    v = np.zeros((0,2))
    for coll in cs.collections:
        ps = coll.get_paths()
        viE = np.zeros((0,2))
        for p in ps:
            viE = np.vstack((viE,p.vertices))
            #print(viE.shape)
        v = np.vstack((v,viE))
    return v

def get_lims(ax,mg,xylims=None,is_3d=0):
    if not xylims:
        if is_3d :
            xmin,xmax = ax.get_xlim3d()
            ymin,ymax = ax.get_ylim3d()
            zmin,zmax = ax.get_zlim3d()
            Wx,Wy,Wz = xmax-xmin,ymax-ymin,zmax-zmin
            xmin,xmax = xmin-mg*Wx, xmax+mg*Wx
            ymin,ymax = ymin-mg*Wy, ymax+mg*Wy
            zmin,zmax = zmin-mg*Wz, zmax+mg*Wz
            xylims = [xmin,xmax,ymin,ymax,zmin,zmax]
        else:
            # data = np.array(ax.dataLim.bounds,dtype=float)
            # data[data==np.inf]=1;data[data==-np.inf]=-1;data[data==np.nan]=0;
            xmin,xmax = ax.get_xlim()
            ymin,ymax = ax.get_ylim()
            W,H = xmax-xmin,ymax-ymin
            xmin,xmax = xmin-mg*W, xmax+mg*W
            ymin,ymax = ymin-mg*H, ymax+mg*H
            xylims = [xmin,xmax,ymin,ymax]
    return xylims

def changeAxesLim(ax,mg,xylims=[],is_3d=0,changeXYlims=0):
    if isinstance(xylims,np.ndarray):xylims=xylims.tolist()
    xylims = get_lims(ax,mg,xylims,is_3d)
    if len(xylims)==4:
        xmin,xmax,ymin,ymax = xylims
    elif len(xylims)==6:
        xmin,xmax,ymin,ymax,zmin,zmax = xylims
    elif len(xylims)==3:
        xylims0 = get_lims(ax,mg,None,is_3d)
        if xylims[0]=='x':
            xmin,xmax = xylims[1:3]
            ymin,ymax = xylims0[2:]
        if xylims[0]=='y':
            xmin,xmax = xylims0[:2]
            ymin,ymax = xylims[1:3]
    if changeXYlims:
        # print([xmin,xmax,ymin,ymax])
        ax.set_xlim((xmin, xmax))
        ax.set_ylim((ymin, ymax))
    xylims = [xmin,xmax,ymin,ymax]
    if is_3d :
        if changeXYlims:ax.set_zlim3d((zmin, zmax))
        xylims +=[zmin, zmax]
    return xylims

def get_tick_array(ticks,ndim,xylims):
    if isinstance(ticks,float) or isinstance(ticks,int) : ticks = [ticks]*ndim
    for i in range(ndim):
        if isinstance(ticks[i],float) or isinstance(ticks[i],int):
            ticks[i]=np.arange(xylims[2*i+0],xylims[2*i+1],ticks[i])
    return ticks
def change_ticks(ax,ticks,tick_labs,xylims,is_3d,ticks_m):
    ndim = [2,3][is_3d]
    if ticks :
        ticks = get_tick_array(ticks,ndim,xylims)
        ax.set_xticks(ticks[0])
        ax.set_yticks(ticks[1])
        if is_3d : ax.set_zticks(ticks[2])
    if ticks_m :
        ticks_m = get_tick_array(ticks_m,ndim,xylims)
        ax.set_xticks(ticks_m[0],minor=True)
        ax.set_yticks(ticks_m[1],minor=True)
    if tick_labs:
        ax.set_xticklabels(tick_labs[0])
        ax.set_yticklabels(tick_labs[1])
        if is_3d : ax.set_yticklabels(tick_labs[2])

# def add_cb(ax):
#     sm = plt.cm.ScalarMappable(cmap=plt.get_cmap('Greys',ns));sm.set_array([]);
#     cb=fig.colorbar(sm,boundaries=0.5+np.arange(ns+1),ticks=range(1,ns+1));
#     cb.ax.set_yticklabels(['%d' %(n) for n in Ns])

def get_axPos(axPosI):
    ''' Positions predefined
    axPos = {'T':[0.2,0.12,0.75,0.75],
        'V':[0.1, 0.1, 0.75, 0.8],
        'L':[0.02, 0.1, 0.75, 0.8],
        1:[0.15, 0.11, 0.82, 0.82],
        11:[0.1, 0.1, 0.35, 0.8],
        12:[0.6, 0.1, 0.35, 0.8],
        21:[0.07, 0.1, 0.9, 0.4],
        22:[0.07, 0.6, 0.9, 0.4],
        31:[0.15, 0.18, 0.75, 0.75]}
    '''
    axPos = {'T':[0.2,0.12,0.75,0.75],
        'V':[0.1, 0.1, 0.75, 0.8],
        'E':[0.1, 0.1, 0.75, 0.75],
        'L':[0.02, 0.1, 0.75, 0.8],
        1:[0.15, 0.11, 0.82, 0.82],
        11:[0.1, 0.1, 0.35, 0.8],
        12:[0.6, 0.1, 0.35, 0.8],
        21:[0.07, 0.1, 0.9, 0.4],
        22:[0.07, 0.6, 0.9, 0.4],
        31:[0.15, 0.18, 0.75, 0.75]}
    axPosition = axPos[1]
    if isinstance(axPosI,list) or isinstance(axPosI,tuple):
        if len(axPosI)==4 : axPosition = axPosI
    elif axPosI in axPos.keys():
        axPosition = axPos[axPosI]
    return axPosition

############
def basename(file):
    if file[-1]=='/' : file = file[:-1]
    return os.path.basename(file)
def dirname(file):
    #if file[-1]=='/' : file = file[:-1]
    return os.path.dirname(file)+'/'

def warn(msg):
    print(colors.red+msg+colors.black)

def get_figpath(file,rel_path):
    figpath=os.path.realpath(os.path.dirname(os.path.realpath(file))+rel_path)+'/'
    return figpath

def saveFig(fullpath,ax=None,png=None,fmt='',figopt='1',topt=0):
    ''' save figure with transparent option
    - figopt : t(transparent) i(quality dpi=i*96)
        - example : 't','1','2', 't1','t2',...
    - topt : bool - transparent option
    '''
    if '.' in fullpath:
        filename,fmt=fullpath,fullpath.split('.')[-1]
    else:
        if not png==None : fmt=['eps','svg','png'][png]
        filename = fullpath+'.'+fmt
    if 't' in figopt : figopt,topt=figopt[1:],1
    r_dpi= 1 if not figopt else int(figopt); #print(r_dpi)
    plt.savefig(filename, format=fmt, dpi=r_dpi*96,transparent=topt)
    print(colors.green+'Saving figure :\n'+colors.yellow+filename+colors.black)

def crop_fig(name,crop):
        cropcmd = "%dx%d+%d+%d" %tuple(crop)
        cmd = "convert %s -crop %s %s" %(name,cropcmd,name)
        check_output(cmd,shell=True)
        print(colors.yellow+name+colors.blue+' cropped to '+colors.black,cropcmd)

def disp_quick(name,ax,opt,figopt):
    if 's' in opt : saveFig(name,ax,fmt='png',figopt=figopt)
    if 'p' in opt : ax.figure.show()
    if 'c' in opt : plt.close(ax.figure)

def im2gif(figpattern,fmt='svg'):
    cmd='im2gif '+figpattern +' ' + fmt
    print(colors.magenta+cmd+'  ...'+colors.black)
    out=check_output(['/bin/bash','-i','-c',cmd]).decode()
    print(colors.green+out+colors.black)

###########################################################################
#misc
def getSymCurve(x,y,symOpt=1):
    x = np.concatenate((np.flipud(-x),x))
    y = np.concatenate((np.flipud(symOpt*y),y))
    return x,y



import easygui

url = 'http://localhost:8002/scattering/linear_array_qdot_sphere/'
help_msg='''
                ###### hard sphere scattering ########
This is a gui for the scattering of a linear array of spheres with constant potential.
More info on the formulation of the problem at :
%s

It is running with matplotlib, so be gentle with it.
Use the SPACEBAR to refresh the view if something goes wrong especially after resizing the window.

## description
In this default setup the problem has been solved for N=2 spheres at few normalized radius ka
and normalized distance kd.

##Graph description :
top-left     : The total scattering cross section
top-right    : The scattering amplitude/differential cross section.The checkboxes on the left side allow you to select the real,imaginary part,phase,magnitude or magnitude square.
bottom-left  : The spherical expansion coefficients in the complex plane for the selected value of
bottom-right : the spheres setup. click on the blue button to display the near field

## GUI usage:''' %url

help_cmd = '''
Click on a (ka,kd) point on the scattering cross section to show the data corresponding to that point
    keyboard commands :
arrows    : move the active selected (ka,kd) horizontally or vertically
enter     : change settings
F1        : show this help
F5        : refresh view
space     : refresh
'''

def multenterbox(msg,title,fieldValues,fieldNames):
    fieldValues = easygui.multenterbox(msg, title, fieldNames,fieldValues)
    while True:
        if fieldValues is None:
            break
        errs = list()
        for n, v in zip(fieldNames, fieldValues):
            if v.strip() == "":errs.append('"{}" is a required field.'.format(n))
        if not len(errs):
            break
        fieldValues = easygui.multenterbox("\n".join(errs), title, fieldNames, fieldValues)
    return fieldValues



# import importlib as imp
import time
import numpy as np
import scipy.special as spe
import py3nj
from . import displayStandards as dsp


w3j = lambda n1,n2,n3,m1,m2,m3 : py3nj.wigner3j(2*n1,2*n2,2*n3,2*m1,2*m2,2*m3)
gaunt = lambda l1,l2,l3,m1,m2,m3:np.sqrt((2*l1+1)*(2*l2+1)*(2*l3+1)/(4*np.pi))*w3j(l1,l2,l3,0,0,0)*w3j(l1,l2,l3,m1,m2,m3)
yn   = lambda n,z : np.sqrt(np.pi/(2*z))*spe.yv(n+0.5,z)
jn   = lambda n,z : np.sqrt(np.pi/(2*z))*spe.jv(n+0.5,z)
hn1  = lambda n,z : np.sqrt(np.pi/(2*z))*spe.hankel1(n+0.5,z)
jnp  = lambda n,z : np.sqrt(np.pi/(2*z))*(spe.jvp(n+0.5,z)  - spe.jv(n+0.5,z)/(2*z) )
hn1p = lambda n,z : np.sqrt(np.pi/(2*z))*(spe.h1vp(n+0.5,z) - spe.hankel1(n+0.5,z)/(2*z) )
Pl = lambda l,theta: np.sqrt(4*np.pi/(2*l+1))*spe.sph_harm(0,l,0,theta)
# psi_lm = lambda z_l,l,m,r,theta,phi:z_l(l,r)*spe.sph_harm(m,l,phi,theta)

def get_MNnm(kr,theta,phi, Nmax, zn=hn1,znp=hn1p,pol='y'):
    '''Computes Mnm and Nmn in spherical coordinates up to Nmax'''
    Plms,dtPlm = get_jmdtPlm(Nmax,theta,phi,phi_opt=False,Ymn_opt=False)
    st = np.sin(theta)
    Ns = dtPlm.shape[0]
    ss = (Ns,3)+kr.shape
    Mnm,Nnm = np.zeros(ss,dtype=complex),np.zeros(ss,dtype=complex)
    kr[kr==0]=1e-3
    if pol=='xy':
        for n in range(1,Nmax+1):
            zn1kr  = zn(n,kr)           #;print(hn1kr)
            dzn1kr = znp(n,kr)+zn1kr/kr #;print(dhn1kr)
            for m in range(n+1):
                idm = int(n*(n+1)/2)+abs(m)
                ep = np.exp(1J*m*phi)   #; print(ep)
                Mnm[idm,1] =  zn1kr*1J*m*Plms[idm]*ep
                Mnm[idm,2] = -zn1kr*dtPlm[idm]*ep
                Nnm[idm,0] =  n*(n+1)*zn1kr/kr*Plms[idm]*st*ep #does not work for m=0 !!!!!!
                Nnm[idm,1] =  dzn1kr*dtPlm[idm]*ep
                Nnm[idm,2] =  dzn1kr*1J*m*Plms[idm]*ep
    #### wiki signs Mo,Ne - x polarization
    if pol=='x':
        for n in range(1,Nmax+1):
            zn1kr  = zn(n,kr)           #;print(hn1kr)
            dzn1kr = znp(n,kr)+zn1kr/kr #;print(dhn1kr)
            for m in range(n+1):
                idm = int(n*(n+1)/2)+abs(m)
                cp,sp = np.cos(m*phi), np.sin(m*phi)  #;print(cp)
                Mnm[idm,1] =  m*cp*Plms[idm]*zn1kr
                Mnm[idm,2] = -sp*dtPlm[idm]*zn1kr
                Nnm[idm,0] =  zn1kr/kr*cp*n*(n+1)*Plms[idm]*st #does not work for m=0 !!!!!!
                Nnm[idm,1] =  cp*dtPlm[idm]*dzn1kr
                Nnm[idm,2] = -m*sp*Plms[idm]*dzn1kr
    ### wiki signs Me,No -y polarization
    if pol=='y':
        for n in range(1,Nmax+1):
            zn1kr  = zn(n,kr)           #;print(hn1kr)
            dzn1kr = znp(n,kr)+zn1kr/kr #;print(dhn1kr)
            for m in range(n+1):
                idm = int(n*(n+1)/2)+abs(m)
                cp,sp = np.cos(m*phi), np.sin(m*phi)  #;print(cp)
                Mnm[idm,1] = -m*sp*Plms[idm]*zn1kr
                Mnm[idm,2] = -cp*dtPlm[idm]*zn1kr
                Nnm[idm,0] =  zn1kr/kr*sp*n*(n+1)*Plms[idm]*st #does not work for m=0 !!!!!!
                Nnm[idm,1] =  sp*dtPlm[idm]*dzn1kr
                Nnm[idm,2] =  m*cp*Plms[idm]*dzn1kr

    return Mnm,Nnm

def get_jmdtPlm(l,theta,phi=None,Ymn_opt=False,phi_opt=False,v=0):
    r'''Computes Plm/sin(cos(theta)) and d_theta Plm(cos(theta)) for :
    - l     : order so that Plm for Plmsin = [Pl00,Pl10,Pl11,Pl20,Pl21,Pl22,Pl30,...]/sin(theta)
    - theta : 2d-array
    - phi,phi_opt,Ymn_opt : obsolete
    returns :
        - Plmsin : $P_l^m(\cos\theta)/\sin\theta$
        - dtPlm  : $\Partial_{\theta}P_l^m(\cos\theta)$
    '''
    nls = int((l+1)*(l+2)/2)    #;print(nls)
    ct,st = np.cos(theta),np.sin(theta)
    if isinstance(theta, int) or isinstance(theta, float):
        ones,zeros,ss = 1,0,(nls)
    else:
        ones,zeros,ss = np.ones(ct.shape,dtype=float),np.zeros(ct.shape,dtype=float),((nls,)+theta.shape)

    if v:print('...Plm/sin...')
    Plmsin = np.zeros(ss,dtype=complex)
    Plmsin[0] = zeros
    Plmsin[1] = zeros
    Plmsin[2] = -ones
    for n in range(2,l+1):
        idn0 = int((n-2)*(n-1)/2)
        idn1 = int((n-1)*(n+0)/2)
        idn2 = int((n-0)*(n+1)/2)
        Plmsin[idn2] = zeros
        for p in range(1,n-1):
            Plmsin[idn2+p] = ((2*(n-1)+1)*ct*Plmsin[idn1+p] - (n-1+p)*Plmsin[idn0+p])/(n-1-p+1); # print('idnp:',idn2+p)
        Plmsin[idn2+n-1] = (2*(n-1)+1)*ct*Plmsin[idn1+n-1]      #; print('idnp:',idn2+n-1, 'using ', idn1+n-1)
        Plmsin[idn2+n] = -(2*(n-1)+1)*st*Plmsin[idn1+n-1]       #;print('idnp:',idn2+n, 'using ', idn1+n)

    if v:print('...derivative...')
    dtPlm  = np.zeros(ss,dtype=complex)
    for n in range(1,l+1):
        idn = int(n*(n+1)/2)
        for p in range(n):
            dtPlm[idn+p] = p*ct*Plmsin[idn+p] + st*Plmsin[idn+p+1]  #;print('idnp:',idn+p)
        dtPlm[idn+n] = n*ct*Plmsin[idn+n]                           #;print('idnp:',idn+n)

    if v:print('...u,v...')
    if phi_opt:
        print('warning : Trying to use Nlm with obsolete phi_opt')
        # for n in range(1,l+1):
        #     idn = int(n*(n+1)/2)
        #     for p in range(n+1):
        #         snm = 1#[1,np.sqrt((2*l+1)/(4*np.pi)*np.product(np.arange(1,l-p+1))/np.product(np.arange(1,l+p+1)))][Ymn_opt]
        #         # print('idnp:',idn+p)
        #         ep = np.exp(1J*p*phi)
        #         Plmsin[idn+p]   *= snm*1J*p*ep
        #         dtPlm[idn+p]    *= snm*ep
    # else:
    #     # print('here')
    #     for n in range(1,l+1):
    #         idn = int(n*(n+1)/2)
    #         for p in range(n+1):
    #             Plmsin[idn+p] *= p

    return Plmsin,dtPlm


def get_MNn1_t(kdp,n,kr,theta,phi,Nmax,zn=jn,znp=jnp,pol='y',fz=np.real):
    Annu,Bnnu = get_vec_trans_coeff_lin0(kdp,n,Nmax)
    Mn1,Nn1   = get_MNn1(kr,theta,phi, Nmax, zn,znp,pol)
    ss = (3,)+kr.shape
    Mr,Nr = np.zeros(ss,dtype=complex),np.zeros(ss,dtype=complex)
    em = 1J#{}
    for n in range(Nmax):
        # Ql1*Mlm[idlm] + im*Pl1*Nlm[idlm]
        Mr +=    Annu[n]*Mn1[n]+em*Bnnu[n]*Nn1[n]
        Nr += em*Annu[n]*Nn1[n]+   Bnnu[n]*Mn1[n]
    if fz:
        return fz(Mr),fz(Nr)
    else:
        return Mr,Nr

def get_MNn1(kr,theta,phi, Nmax, zn=hn1,znp=hn1p,pol='y'):
    '''Computes Mn1 and Nm1 in spherical coordinates up to Nmax'''

    Pl1s,dtPl1 = get_Pl1sdtPl1(Nmax,theta)
    st    = np.sin(theta)
    cp,sp = np.cos(phi), np.sin(phi)  #;print(cp)

    Ns = dtPl1.shape[0]
    ss = (Ns,3)+kr.shape
    kr[kr==0]=1e-3
    Mn1,Nn1 = np.zeros(ss,dtype=complex),np.zeros(ss,dtype=complex)
    m=1
    #### wiki signs Mo,Ne - x polarization
    if pol=='x':
        for n in range(1,Nmax+1):
            zn1kr  = zn(n,kr)
            dzn1kr = znp(n,kr)+zn1kr/kr
            Mnm[n-1,1] =  m*cp*Pl1s[n]*zn1kr
            Mnm[n-1,2] = -sp*dtPl1[n]*zn1kr
            Nnm[n-1,0] =  zn1kr/kr*cp*n*(n+1)*Pl1s[n]*st
            Nnm[n-1,1] =  cp*dtPl1[n]*dzn1kr
            Nnm[n-1,2] = -m*sp*Pl1s[n]*dzn1kr
    ### wiki signs Me,No -y polarization
    if pol=='y':
        for n in range(1,Nmax+1):
            zn1kr  = zn(n,kr)
            dzn1kr = znp(n,kr)+zn1kr/kr
            Mn1[n-1,1] = -m*sp*Pl1s[n]*zn1kr
            Mn1[n-1,2] = -cp*dtPl1[n]*zn1kr
            Nn1[n-1,0] =  zn1kr/kr*sp*n*(n+1)*Pl1s[n]*st
            Nn1[n-1,1] =  sp*dtPl1[n]*dzn1kr
            Nn1[n-1,2] =  m*cp*Pl1s[n]*dzn1kr

    return Mn1,Nn1

def get_Pl1sdtPl1(l,theta,v=0):
    r'''Computes Pl1/sin(cos(theta)) and dP_theta Pl1(cos(theta)) for :
    - l     : order so that Pl1 for Pl1sin = [Pl00,Pl10,Pl20,Pl30,...]/sin(theta)
    - theta : 2d-array
    - phi,phi_opt,Ymn_opt : obsolete
    returns :
        - Plmsin : $P_l^1(\cos\theta)/\sin\theta$
        - dtPlm  : $\Partial_{\theta}P_l^1(\cos\theta)$
    '''
    nls = l+1
    ct,st = np.cos(theta),np.sin(theta)
    if isinstance(theta, int) or isinstance(theta, float):
        ones,zeros,ss = 1,0,(nls)
    else:
        ones,zeros,ss = np.ones(ct.shape,dtype=float),np.zeros(ct.shape,dtype=float),((nls,)+theta.shape)

    if v:print('...Plm/sin...')
    Pl1s = np.zeros(ss,dtype=complex)
    Pl1s[1] = -ones
    for n in range(1,l):
        Pl1s[n+1] = ((2*n+1)*ct*Pl1s[n] - (n+1)*Pl1s[n-1])/n; # print('idnp:',idn2+p)

    if v:print('...derivative...')
    dtPl1  = np.zeros(ss,dtype=complex)
    dtPl1[1] = ct*Pl1s[1]
    for n in range(2,l+1):
        dtPl1[n] = n*ct*Pl1s[n] - (n+1)*Pl1s[n-1]  #;print('idnp:',idn+p)

    return Pl1s,dtPl1

################################################################################
#                           Plane wave
################################################################################
def get_plane_wave_scalar(y,z,lam=1,d_p=2.5,alpha=0,Nmax=10,fz=np.real,timeOpt=False):
    r'''Plane Wave Ex = exp(jkz) expansion upon Spherical waves using :
            $$e^{jkz} = e^{jkr\cos(\theta)}=sum_{n=0}^{\infty} a_nJ_n(kr)P_n(\cos(\theta))$$
    - lam : wavelength
    - d_p : distance to origin
    - alpha : incident angle degrees
    - Nmax  : max order expansion
    -fz     : np.real,np.imag,np.abs or None
    '''
    k,alpha,npts = 2*np.pi/lam, np.deg2rad(alpha),y.shape[0]
    r,theta = np.sqrt(y**2+z**2),np.arctan(y/z)
    Ex = np.zeros(z.shape,dtype=complex)
    n  = np.arange(Nmax+1)
    if alpha:
        t = time.time()
        for n in range(Nmax+1):
            ymn = np.zeros(Ex.shape,dtype=complex)
            for m in range(-n,n+1):
                ymn += spe.sph_harm(m,n,0,theta)*np.conj(spe.sph_harm(m,n,0,alpha))
            Ex += 1J**n*spe.spherical_jn(n,k*r)*ymn
        if timeOpt:print('loop:',time.time()-t)
        Ex *= 4*np.pi*np.exp(1J*k*d_p*np.cos(alpha))
    else:
        t = time.time()
        Pn = np.zeros((npts,npts,Nmax+1))
        for i in range(npts): #enumerate(np.cos(theta.flatten()):
            for j in range(npts):
                Pn[i,j,:] = spe.lpn(Nmax,np.cos(theta[i,j]))[0]
        if timeOpt:print('poly:',time.time()-t)

        t=time.time()
        Jn = spe.spherical_jn
        an = lambda n:1J**(-n)*(2*n+1)
        for n in range(Nmax+1):
            Ex += an(n)*Jn(n,k*r)*Pn[:,:,n]
        if timeOpt:print('sum Jn:',time.time()-t)
    if fz:
        return fz(Ex)
    else:
         return Ex


def get_plane_wave_vector(r,theta,phi=np.pi/2,lam=1,alpha=0,d_p=0,Nmax=10,fz=np.real,pol='x',v=False):
    r'''Plane Wave expansion upon Vector Spherical waves
        - r,theta,phi : grid coordinates
        - lam : wavelength
        - d_p : distance to origin
        - alpha : incident angle (degrees)
        - Nmax  : max order expansion (inclusive)
        -fz     : np.real,np.imag,np.abs or None
    returns :
        - E = Er,Et,Ep for plane wave
    '''
    k,alpha = 2*np.pi/lam, np.deg2rad(alpha)

    if v:print('\t\t plane wave vector')
    E = np.zeros((3,)+r.shape,dtype=complex)

    if alpha>0:
        if v:print('...Mlm,Nlm,Plm(alpha)... : ',end='');t = time.time()
        Mlm,Nlm = get_MNnm(k*r,theta,phi, Nmax, zn=jn, znp=jnp,pol=pol)
        jmPlm,dtPlm = get_jmdtPlm(Nmax,alpha,phi=phi,Ymn_opt=False,v=0)
        if v:print('%.2f'%(time.time()-t))

        if v:print('..summation.. : ',end='');t = time.time()
        for l in range(1,Nmax+1):
            for m in range(-l,l+1):
                slm = np.product(np.arange(1,l-abs(m)+1))/np.product(np.arange(1,l+abs(m)+1))
                idlm = int(l*(l+1)/2)+abs(m)
                Plm = -1J**l*(2*l+1)/(l*(l+1))*slm*jmPlm[idlm]
                Qlm = -1J**l*(2*l+1)/(l*(l+1))*slm*dtPlm[idlm]
                # print('l=%-3d,m=%-3d, slm=%.1E,' %(l,m,slm))
                # print('\t P=%.1E, Q=%.1E' %(Plm,Qlm))#, jmPlm[idlm.max(),dtPlm[idlm].max()))
                E += Plm*Nlm[idlm]+Qlm*Mlm[idlm]
    else:
        # Ml1,Nl1 = get_MNn1(k*r,theta,phi, Nmax, zn=jn, znp=jnp)
        Mlm,Nlm = get_MNnm(k*r,theta,phi, Nmax, zn=jn, znp=jnp,pol=pol)
        if v:print('..summation.. : ',end='');t = time.time()
        im = {'x':-1J,'y':1J,'xy':1}[pol]
        for l in range(1,Nmax+1):
            idlm = int(l*(l+1)/2)+1
            Pl1 = -1J**l*(2*l+1)/(2*l*(l+1))
            Ql1 = -1J**l*(2*l+1)/(2*l*(l+1))
            # print('l=%-3d,m=%-3d, slm=%.1E,' %(l,m,slm))
            # print('\t P=%.1E, Q=%.1E' %(Plm,Qlm))#, jmPlm[idlm.max(),dtPlm[idlm].max()))
            E +=  Ql1*Mlm[idlm] + im*Pl1*Nlm[idlm]
        if d_p:E *= np.exp(1J*k*d_p)

    if v:print('%.2f'%(time.time()-t))

    if fz:
        return fz(E)
    else:
         return E


########################################################################################
#                               Translation addition theorem
########################################################################################
# def get_vec_trans_coeff(l,m,n,p,**kwargs):
#     '''vectorial translational addition theorem coefficients'''
#     Almnp = 1/(n*(n+1))*(
#         0.5*np.sqrt((l-m)*(l+m+1)*(n-p)*(n+p+1))*a_lmnp(l,m+1,n,p+1,**kwargs)+
#         0.5*np.sqrt((l+m)*(l-m+1)*(n+p)*(n-p+1))*a_lmnp(l,m-1,n,p-1,**kwargs)+
#         m*p*a_lmnp(l,m,n,p,**kwargs)
#         )
#     Blmnp =(
#     -0.5*np.sqrt((l-m)*(l+m+1))*(
#         1/(n+1)*np.sqrt((n+p+1)*(n+p+2)/((2*n+1)*(2*n+3)))*a_lmnp(l,m+1,n+1,p+1,**kwargs)+
#         1/n    *np.sqrt((n-p-1)*(n-p  )/((2*n-1)*(2*n+1)))*a_lmnp(l,m+1,n-1,p+1,**kwargs)
#         )
#     +0.5*np.sqrt((l+m)*(l-m+1))*(
#         1/(n+1)*np.sqrt((n-p+1)*(n-p+2)/((2*n+1)*(2*n+3)))*a_lmnp(l,m-1,n+1,p-1,**kwargs)+
#         1/n    *np.sqrt((n+p-1)*(n+p  )/((2*n-1)*(2*n+1)))*a_lmnp(l,m-1,n-1,p-1,**kwargs)
#         )
#     +m*(
#         1/(n+1)*np.sqrt((n+p+1)*(n-p+1)/((2*n+1)*(2*n+3)))*a_lmnp(l,m,n+1,p,**kwargs)+
#         -1/n   *np.sqrt((n+p  )*(n-p  )/((2*n-1)*(2*n+1)))*a_lmnp(l,m,n-1,p,**kwargs)
#         )
#     )
#     return Almnp,Blmnp
#

def a_ln(lmax,nmax, fz_q,r_d,theta_d):
    '''Scalar translational addition theorem coefficients
    - Nmax : order of expansion
    - fz_q : type of expansion (bessel or hankel)
    - r_d,theta_d : translation vector
    '''
    l = np.arange(lmax+nmax+2)
    Yq  = np.sqrt((2*l+1)/(4*np.pi))*spe.lpn(lmax+nmax+1,np.cos(theta_d))[0]
    # Yq = np.array([spe.sph_harm(0,q,0,theta_d) for q in range(lmax+nmax)])
    zq = np.array([fz_q(q,r_d) for q in range(lmax+nmax+2)])

    aln = np.zeros((lmax+1,nmax+1),dtype=complex)
    for l in range(lmax+1):
        # print('l=%d' %l)
        for n in range(nmax+1):
            q = np.arange(abs(l-n),n+l+1)   #;cput = time.time()
            Glnq = np.array([np.float(gaunt(l,n,q, 0,0,0)) for q in q])
            # Glnq = spu.w3j(l,n,q, 0,0,0)
            aln[l,n] = 4*np.pi*np.sum((-1J)**(l-n-q)*zq[q]*Yq[q]*Glnq) #;print('n=%d' %n, q,zq[q],Yq[q],Glnq)
    return aln

def a_lmnp(l,m,n,p, fz_q,r_d,theta_d,phi_d):
    '''Scalar translational addition theorem coefficients
    - l,m,n,p : index to pass from psi_np to psi_lm
    - fz_q : type of expansion (bessel or hankel)
    '''
    q = np.arange(abs(l-n),n+l+1)
    cput = time.time()
    # Glmnpq = np.array([np.float(gaunt(l,n,q, m,p,m-p)) for q in q])                           #;printf('Glmnpq:',t-cput) = time.time()
    Glmnpq = np.array([np.float(gaunt(l,n,q, m,-p,-m+p)) for q in q])                           #;printf('Glmnpq:',t-cput) = time.time()
    z_q    = np.array([fz_q(q,r_d) for q in q])                                                 #;printf('zq    :',t-cput) = time.time()
    Ympq   = np.array([ [0,spe.sph_harm(m-p,q,phi_d,theta_d)][bool(abs(m-p)<=q)] for q in q ])  #;printf('Ympq  :',t-cput) = time.time()
    # a_lmnp = 4*np.pi*np.sum(
    #     (-1J)**(n-l-q)*z_q*Ympq*Glmnpq*(-1)**m)
    a_lmnp = 4*np.pi*np.sum(
        (-1J)**(l-n-q)*z_q*Ympq*Glmnpq*(-1)**m)
    return a_lmnp

def get_trans_coeff(n,p,r_d,theta_d,phi_d, N=5):
    a_lmnp,i = np.zeros((N**2),dtype=complex),0
    for l in range(N):
        q = np.arange(abs(l-n),n+l+1)
        Jq = np.array([spe.spherical_jn(q,r_d) for q in q])
        # print('l=%-3d' %l,q,Jq )
        for m in range(-l,l+1):
            Glmnpq = np.array([np.float(gaunt(l,n,q, m,-p,-m+p)) for q in q])
            Ympq = np.array([ [0,spe.sph_harm(m-p,q,phi_d,theta_d)][bool(abs(m-p)<=q)] for q in q ])
            # print('\t m=%-3d : ' %m,Glmnpq,Ympq)
            a_lmnp[i] = 4*np.pi*np.sum(
                (-1J)**(l-n-q)*(-1)**m*Glmnpq*Jq*Ympq)
            i+=1
    return a_lmnp

def translation_addth_scalar(r,theta,phi, l,m,r_d,theta_d,phi_d,Nmax=10,zl=jn,zq=jn):
    '''Scalar translation addition theorem for psi_lm
    - r,theta,phi       : coordinates at which output is calculated
    - r_d,theta_d,phi_d : centre of spherical coordinates reference
    - l,m   : index of output function
    - Nmax  : max order of expansion
    - zl : function of the expansion jn('in') or hn1('out')
    - zq : function jn(out-out or in-in) or hn1(out-in or in-out)
    returns:
    - psi_lm : psi_lm(r,theta,phi) = almnp(r_d,theta_d,phi_d)*psi_np(r,theta,phi)
    '''
    # print(r_d,theta_d,phi_d,Nmax)
    ns = np.hstack([ [n]*(2*n+1) for n in np.arange(Nmax)])
    ps = np.hstack([np.arange(-n,n+1) for n in np.arange(Nmax)])
    psi_lm = np.zeros(r.shape,dtype=complex)
    a=[]
    for n,p in zip(ns,ps):
        almnp = a_lmnp(l,m,n,p, zq,r_d,theta_d,phi_d)
        psi_lm += almnp*zl(n,r)*spe.sph_harm(p,n,phi,theta) #;print(n,p,almnp)
        a+=[almnp]
    a=np.abs(np.array(a));ida=a>0;a=a[ida]; print(a,ns[ida],ps[ida])
    # print(['%.2E' %i.real for i in a])

    # almnp = get_trans_coeff(l,m,r_d,theta_d,phi_d, N=Nmax)
    # for i in range(ns.size):
    #     psi_lm += almnp[i]*zl(ns[i],r)*spe.sph_harm(ps[i],ns[i],phi,theta)

    # print(['%.2E' %i for i in almnp.real])
    return psi_lm



def get_vec_trans_coeff_lin0(kdp,n,Nmax=5):
    '''vectorial translational addition coefficients for endfired (normal incidence alpha=0) linear configuration
        - kdp  : normalized translation distance
        - n    : int - order(n>=1) of function to translate
        - Nmax : max included order
    returns :
        - Annu,Bnnu : coefficients nu=1..Nmax with m=mu=1
    '''
    nus = np.arange(Nmax)+1
    hp = np.array([hn1(p, kdp) for p in range(n+Nmax)])
    Annu,Bnnu = np.zeros((Nmax),dtype=complex),np.zeros((Nmax),dtype=complex)
    for nu in nus:
        p = np.arange(abs(n-nu),n+nu)
        Annup,Annupq = (2*p+1)*np.sqrt(n*(n+1)/(nu*(nu+1)))*w3j(n,nu,p,1,-1,0)*np.array([w3j(n,nu,p,0,0,0),w3j(n,nu,np.maximum(p-1,0),0,0,0)])
        annup = (1J)**(nu-n+p)/(2*nu*(nu+1))*(
            2*nu*(nu+1)*(2*nu+1) + (nu+1)*(n+nu-p)*(n+p-nu+1) - nu*(n+nu+p+2)*(nu+p-n+1))
        bnnup = -(1J)**(nu-n+p)*(2*nu+1)/(2*nu*(nu+1))*np.sqrt(
            (n+nu+p+1)*(nu+p-n)*(n+p-nu)*(n+nu-p+1))

        Annu[nu-1] = np.sum( Annup *annup*hp[p])
        Bnnu[nu-1] = np.sum(-Annupq*bnnup*hp[p])

        #f=lambda x:np.abs(x).max()
        #print('n=%d, nu=%d, p =' %(n,nu), p);print('hp=%.1f,Anp=%.1f,anp=%.1f,Anq=%.1f,bnp=%.1f' %(f(hp),f(Annup),f(annup),f(Annupq),f(bnnup)))

    return Annu,Bnnu


##################################################################################################
#                   MESH, COORDS
##################################################################################################
def sphere2cart(r,theta,phi):
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return x,y,z

def cart2sphere(x,y,z):
    r   = np.sqrt(x*x+y*y+z*z)
    rho = np.sqrt(x**2+y**2)
    theta = np.arccos(z/r)
    phi   = np.arctan2(y,x)
    return r,theta,phi

def Esphere2cart(Esph,theta,phi):
    ''' Convert field map from spherical coordinates to cartesian
    - Esp : spherical field map
    - theta,phi : angular spherical coordinates
    returns :
    - E : cartesian field map
    '''
    E = np.zeros(Esph.shape)
    ct,st = np.cos(theta),np.sin(theta)
    cp,sp = np.cos(phi),np.sin(phi)

    E[0] =  st*cp*Esph[0]  + ct*cp*Esph[1] - sp*Esph[2] #E_x
    E[1] =  st*sp*Esph[0]  + ct*sp*Esph[1] + cp*Esph[2] #E_y
    E[2] =  ct   *Esph[0]  - st   *Esph[1] + 0 *Esph[2] #E_z
    return E

def polar_mesh2D(cart_opt=False,npts=100,r=(0,1)):
    ''' 2D polar mesh :
        - cart_opt : True(cartesia mesh) ot False(polar mesh)
        - r     :
            - cart_opt=True  (4-tuple) - y[0],z[0]
            - cart_opt=False (2-tuple) - range of radius
        - npts  : int - discretization
    returns :
        - r,theta,y,z : polar mesh and corresponding cartesian
    '''
    if cart_opt:
        if len(r)==2:r=r*2
        y,z = np.meshgrid( np.linspace(r[0],r[1],npts), np.linspace(r[2],r[3],npts))
        r,theta = np.sqrt(y**2+z**2),np.arctan2(y,z)
    else:
        r,theta = np.meshgrid(np.linspace(r[0],r[1],npts),np.linspace(0,np.pi,npts))
        z,y = r*np.cos(theta), r*np.sin(theta)
    return r,theta,y,z

def mesh_3D(npts=50,r=(0,1)):
    if len(r)==2:r=r*3
    u = np.linspace(r[0],r[1],npts)
    v = np.linspace(r[2],r[3],npts)
    w = np.linspace(r[4],r[5],npts)
    x,y,z = np.meshgrid(u,v,w)
    return x,y,z

##################################################################################################
#                   Display_utils
##################################################################################################
def plot_E3d_plane(fE,cmpts=[0,1,2],cart_opt=True,r=5,npts=50,phi=0,name='E',fieldname='$E$',caxis=None,**kwargs):
    r'''plots 3d field map in phi-planes
    - fE : function to compute E as fE(r,theta,phi)
    - cmpts     : components to plot
    - cart_opt  : 'M'(cartesian mesh), 'E'(Fieldmap)
    - r,npts    : mesh options
    - phi       : phi-plane
    - field : str - field name for title
    - name  : str - fig name prefix - name+'%s.png' %cmpt
    '''
    if isinstance(cart_opt,bool):['','ME'][cart_opt]

    r,theta,y,z = polar_mesh2D('M' in cart_opt,npts,r)
    E = fE(r,theta,phi)

    if 'E' in cart_opt:
        E = Esphere2cart(E,theta,phi)
        scmpt,ncmpt=['x','y','z'],['x','y','z']
    else:
        scmpt,ncmpt = ['r',r'\theta',r'\phi'],['r','t','p']

    if not caxis:
        mm = np.abs([E.min(),E.max()]).max()
        caxis=[-mm,mm]
    #display
    for cmpt in cmpts:
        tle  = 'Field %s \n $e_{%s}$ component' %(fieldname,scmpt[cmpt])
        nameE = '_%s.png' %(ncmpt[cmpt])

        dsp.stddisp(im=[y,z,E[cmpt]],labs=['$y$','$z$'],title=tle,
            imOpt='cv',caxis=caxis,axPos='V',
            name=name+nameE,**kwargs)

class sphere_array:
    def biRCS_show(self,npts=100,is_3d=True,phi=[0,90],bargs={},im='',**kwargs):
        ''' plots bistatic RCS
        - npts  : int - nb points in theta and phi(if is_3d=1)
        - is_3d : bool - True=3d plot
        - phi   : list or np.ndarray - discrete values of phi(degrees) for 2d case(is_3d=0)
        '''
        theta = np.linspace(0,np.pi,2*npts)
        if is_3d:
            phi = np.linspace(0,2*np.pi,npts)
        else:
            phi = np.deg2rad(phi)
        theta3d,phi3d = np.meshgrid(theta,phi)
        rcs = self.compute_biRCS(theta3d,phi3d,**bargs)

        tle = r'bistatic RCS $\lambda$=%.1f, $a_p$=%.1f, $ka=$%.1f  $\alpha$=%d$^{\circ}$, $N=%d$' %(self.lam,self.a_p[0],2*np.pi*self.a_p[0]/self.lam, np.rad2deg(self.alpha),self.N)
        if is_3d:
            x,y,z =sphu.sphere2cart(np.log10(rcs),theta3d,phi3d)
            dsp.stddisp(scat=[x,y,z,np.abs(z)],rc='3d',labs=['$x$','$y$','$z$'],**kwargs)
        else:
            mono = self.compute_monoRCS(**bargs)
            cs   = dsp.getCs('Spectral',phi.size); # print(theta.shape,rcs.shape)
            plts = [[np.rad2deg(theta),rcs[i,:],cs[i],r'$\phi=%d^{\circ}$' %phi] for i,phi in enumerate(np.rad2deg(phi))]
            plts += [[180,mono,'bo' ,'$mono$']]
            labs = [r'$\theta(^{\circ})$',r'$\sigma(\theta,\phi)/\pi a_0^2$']
            if im:
                dsp.image_bg(im,plots=plts,lw=2,labs=labs,title=tle,
                    **kwargs)
            else:
                dsp.stddisp(plts,lw=2,labs=labs,title=tle,name=name,
                    **kwargs)

    def fields_show(self,ftype='t',cmpts=[0,1,2],idp=None,phi=np.pi/2,name='',
        cart_opt=False,npts=200,r=(1,10),**kwargs):
        ''' Plot the near field electric field
        - ftype : str - 't'(total), 's'(scattered), 'i'(incident)
        - cmpts : list- indices of components to plot 0(E_r) 1(E_theta) 2(E_phi)
        - idp   : int - sphere index (None => all)
        - phi   : float - angle(rad) for the phi plane
        - name  : prefix for name of figure where naming convention will be :
            name+'E%s%s_%s.png' %(ftype,cmpt,idp)
        - cart_opt,npts,r : mesh options, see spherical_utils.polar_mesh2D
        '''
        print('...Computing fields...')
        r,theta,y,z = sphu.polar_mesh2D(cart_opt,npts,r)
        E = self.compute_fields(r,theta,phi,ftype,idp)

        print('...Plotting fields...')
        ftypes = {'t':'Total','s':'Scattered', 'i':'Incident'}
        str_p,str_pn='',''
        if isinstance(idp,int):str_p, str_pn=' sphere %d' %(idp), '_%d' %(idp)

        #spheres
        t = np.linspace(-np.pi/2,np.pi/2,100)
        ct,st = np.cos(t),np.sin(t)
        plts = [ [ap*ct, dp+ap*st,'k-',''] for ap,dp in zip(self.a_p,self.d_p)]
        for cmpt in cmpts:
            mm = np.abs([E[cmpt].min(),E[cmpt].max()]).max()
            tle = r'%s Field $E^{%s}_{%s}$ '%(ftypes[ftype],ftype,['r',r'\theta', r'\phi'][cmpt])
            tle+=str_p
            dsp.stddisp(plots=plts,im=[y,z,E[cmpt]],labs=['$y$','$z$'],title=tle,
                imOpt='cv',caxis=[-mm,mm],axPos='V',lw=2,
                name=name+'E%s%s%s.png' %(ftype,['r','t','p'][cmpt],str_pn),**kwargs)
        return y,z,E

    def save(self,file=''):
        if not file:file='PECspheres%d%d.pkl' %(self.N,self.lam)
        with open(file,'wb') as out :
            pickle.dump(self, out, pickle.HIGHEST_PROTOCOL)
        if v:print(colors.green+"object saved\n"+colors.yellow+file+colors.black)


# spe.lmnp
# mnify = lambda x:np.hstack([np.hstack([np.flipud(x[1:n+1,n]),x[:n+1,n]]) for n in range(Nmax+1)])
# m = np.hstack([np.arange(-n,n+1) for n in np.arange(Nmax+1)])
# n = np.hstack([[n]*(2*n+1) for n in np.arange(Nmax+1)])


# if __name__== "__main__":
#     plt.close('all')
    # print(get_jmdtPlm(l=3,theta=0,phi=np.pi,v=0))
    # print(get_jmdtPlm(l=1,theta=np.pi/6))



# import importlib as imp
import scipy.special as spe
import numpy as np
from . import displayStandards as dsp  #;imp.reload(dsp)
from . import spherical_utils as spu  #;imp.reload(spu)
from . import hard_sphere_base as hsb #;imp.reload(hsb)

Pl = lambda l,theta: np.sqrt(4*np.pi/(2*l+1))*spe.sph_harm(0,l,0,theta)

class HardSphereArray(hsb.HardSphereArrayBase):
    def solve(self,copt=1,v=0):
        N,ka,kd,nmax = self.N,self.ka,self.kd,self.nmax
        kdp = self.kd_p

        ll = np.arange(nmax)
        bl = 1J**ll*(2*ll+1)
        ul = -np.array([spu.jn(l,ka)/spu.hn1(l,ka) for l in range(nmax)])
        cpl0 = np.sqrt(4*np.pi/(2*ll+1))*bl*ul
        Yl = np.array([spe.sph_harm(0,l,0,[0,np.pi]) for l in range(nmax)])

        f = np.zeros(2,dtype=complex)#np.exp(1j*kdp)
        for l in range(nmax):
            f += (-1J)**(l+1)*cpl0[l]*Yl[l]

        self.L,self.T = np.zeros((N),dtype=complex),np.zeros((N,N),dtype=complex)
        self.Cp = np.zeros((N),dtype=complex)
        if copt:
            for p in range(self.N):
                fp = f*np.exp(1J*kdp[p])  #;print('p=%d, kdp=%.1f, f_p=' %(p,kdp[p]),fp)

                # print(colors.yellow+'p=%d'%p+colors.black)
                q = np.setdiff1d(np.arange(N),p)
                kdpq = np.abs(kdp[p]-kdp[q])    #; print('f1:',2*f/kdpl)
                theta_qp = np.array(q>p,dtype=int) #0,0,.,pi,pi
                theta_pq = np.array(q<p,dtype=int) #pi,pi,.,0,0
                gp1 = np.sum(fp[theta_pq])/kdpq
                gp2 = np.sum(fp[theta_qp])/kdpq
                # incident field scattering term
                self.L[p] = gp2
                # rear spheres scattering terms
                r  = np.arange(p)
                self.T[p,r] = np.exp(-1J*kdp[r])*gp1  #;print('q:',q)
                # forward spheres scattering terms
                r  = np.arange(p+1,N)
                self.T[p,r] = np.exp(+1J*kdp[r])*gp2  #;print('q:',q)

            if v:print("...solving...")
            self.Cp = np.linalg.solve((N-1)*np.identity(N)-self.T,self.L)
        return self.Cp

    def _cpl(self):
        k,ka,kdp,nmax,N = self.k,self.ka,self.kd_p,self.nmax,self.N
        ll = np.arange(nmax)
        bl = 1J**ll*(2*ll+1)
        ul = -np.array([spu.jn(l,ka)/spu.hn1(l,ka) for l in range(nmax)])
        cpl0 = np.sqrt(4*np.pi/(2*ll+1))*bl*ul
        self.cpl = np.zeros((N*nmax),dtype=complex)
        for l in range(nmax):
            for p in range(self.N):
                q0,q1 = np.arange(p-1),np.arange(p+1,self.N)
                self.cpl[p*nmax+l]=cpl0[l]*np.exp(1J*kdp[p])*(
                    1 + np.sum(self.Cp[q0]*np.exp(-1J*kdp[q0])) +
                    (-1)**l*np.sum(self.Cp[q1]*np.exp(+1J*kdp[q1]))
                    )
        return self.cpl


    def get_ff(self,theta=np.linspace(0,np.pi,361)):
        '''far field scattering amplitude'''
        k,ka,kdp,nmax,N = self.k,self.ka,self.kd_p,self.nmax,self.N
        ct = np.cos(theta)

        ll = np.arange(nmax)
        bl = 1J**ll*(2*ll+1)
        ul = -np.array([spu.jn(l,ka)/spu.hn1(l,ka) for l in range(nmax)])
        cpl0 = np.sqrt(4*np.pi/(2*ll+1))*bl*ul

        f = np.zeros(theta.shape,dtype=complex)
        for l in range(nmax):
            Yl = spe.sph_harm(0,l,0,theta)
            for p in range(self.N):
                q0,q1 = np.arange(p-1),np.arange(p+1,self.N)
                f+=(-1J)**(l+1)*cpl0[l]*np.exp(1J*kdp[p])*(
                    Yl +
                    np.sum(self.Cp[q0]*np.exp(-1J*kdp[q0]))*Yl +
                    np.sum(self.Cp[q1]*np.exp(+1J*kdp[q1]))*np.flipud(Yl)
                    )*np.exp(-1J*kdp[p]*ct)
        return f



import setuptools

with open("README.md", "r") as fh:
   long_description = fh.read()

setuptools.setup(
    name="pyScatSpheres",
    version="1.0.3.dev3",
    author="Tarik Ronan Drevon",
    author_email="tarik.drevon@stfc.ac.uk",
    description="Scattering of array of spheres, scalar theory",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/pyScatSpheres",
    project_urls={
        'Documentation': 'https://pyscatspheres.readthedocs.io/en/latest/',
        'Source':'https://github.com/ronandrevon/pyScatSpheres',
    },
    keywords=["diffraction","spherical","scalar","wave","array"],
    python_requires='>=3.6',
    packages=['pyScatSpheres'],
    include_package_data=True,
    package_data={'':['data*.pkl']},
    install_requires=['numpy','scipy','pandas','colorama','py3nj','matplotlib','easygui'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License ",
        "Operating System :: OS Independent",
    ],
)



# import importlib as imp
import scipy.special as spe
import numpy as np, pandas as pd
from . import displayStandards as dsp #;imp.reload(dsp)
from . import glob_colors as colors
from . import spherical_utils as spu  #;imp.reload(spu)
from . import hard_sphere_base as hsb #;imp.reload(hsb)

class QdotSphereArray(hsb.HardSphereArrayBase):
    def solve(self,nmax=-1,copt=1,opt2=0,v=0):
        ''' Finds the unknown coefficients
        - nmax : max inlcuded order
        - copt : solve coupled problem
        - opt2 : get
        '''
        if nmax>0:self.nmax=nmax+1
        N,n_p,ka,kd,kdp,nmax = self.N,self.kp,self.ka,self.kd,self.kd_p,self.nmax
        fz_q = spu.hn1

        #preprocess,
        ll = np.arange(nmax)
        cl = 1J**ll*(2*ll+1)*np.sqrt(4*np.pi/(2*ll+1))
        jl0,jl1   = np.array([spu.jn( l,np.array([ka,n_p*ka]))  for l in ll]).T
        djl0,djl1 = np.array([spu.jnp(l,np.array([ka,n_p*ka]))  for l in ll]).T
        hl0  = np.array([spu.hn1(l,ka)  for l in ll])
        dhl0 = np.array([spu.hn1p(l,ka) for l in ll])
        if opt2:
            al = cl*(hl0*djl0-    dhl0*jl0)/(n_p*djl1*hl0-jl1*dhl0)
            bl = cl*(jl1*djl0-n_p*djl1*jl0)/(n_p*djl1*hl0-jl1*dhl0)
        else:
            al = cl*jl0
            bl = cl*djl0

        #### incident plane wave response
        L = np.hstack([np.tile(al,N),np.tile(bl,N)])
        for p in range(N):
            idpl = p*nmax+np.arange(nmax) #;print('p=%d'%p)#;print('idp:' , idpl)
            L[idpl]        *= np.exp(1J*kdp[p])
            L[idpl+N*nmax] *= np.exp(1J*kdp[p])
        if opt2:
            self.ap,self.bp = L[:N*nmax],L[N*nmax:]
            return self.ap,self.bp

        #### assembling
        T = np.zeros((2*N*nmax,2*N*nmax),dtype=complex)
        P = np.zeros((2*N*nmax,2*N*nmax),dtype=complex)
        if v:print("...assembling coupling...")
        for p in range(N):
            print(colors.yellow+'p=%d' %p+colors.black)
            idp1,idp2 = slice(p*nmax,(p+1)*nmax),slice((N+p)*nmax,(N+p+1)*nmax)
            P[idp1,idp1] =  np.diag(jl1)
            P[idp1,idp2] = -np.diag(hl0)
            P[idp2,idp1] =  np.diag(djl1*n_p)
            P[idp2,idp2] = -np.diag(dhl0)
            if copt:
                for q in range(p):
                    kdpq = kdp[p]-kdp[q]
                    idqb = slice((N+q)*nmax,(N+q+1)*nmax)
                    aln0 = spu.a_ln(nmax-1,nmax-1, fz_q,kdpq,np.pi)
                    T[idp1,idqb] =  jl0[:,None] * aln0
                    T[idp2,idqb] = djl0[:,None] * aln0
                for q in range(p+1,N):
                    idqb = slice((N+q)*nmax,(N+q+1)*nmax)
                    kdpq = kdp[q]-kdp[p]
                    aln1 = spu.a_ln(nmax-1,nmax-1, fz_q,kdpq,0)
                    T[idp1,idqb] =  jl0[:,None]* aln1
                    T[idp2,idqb] = djl0[:,None]* aln1
        #### Solving
        if v:print("...solving...")
        cp = np.linalg.solve(P-copt*T,L)
        self.ap,self.bp = cp[:N*nmax],cp[N*nmax:]

        return self.ap,self.bp

    def set_Cp(self,cp):
        # print(self.N,cp)
        self.ap,self.bp = cp[:self.N*self.nmax],cp[self.N*self.nmax:]


    def compute_f(self,r,theta,phi,ftype='t',Gopt=0,idp=None):
        ''' computes scattering amplitude f
        - r,theta,phi : np.ndarray each - coordinates
        - ftype : str - 't'(total), 's'(scattered), 'i'(incident),
        - Gopt : compute gradient
        - idp : index of sphere to show (None or -1 => all)
        return :
        - np.real(E_ftype)
        '''
        k,n_p,d_p,nmax,N = self.k,self.kp,self.d_p,self.nmax,self.N
        x,y,z = spu.sphere2cart(r,theta,phi)
        idp = self._check_idp(idp) #;print(idp)

        #incident wave
        if ftype in 'ita':
            fi = np.zeros(r.shape,dtype=complex)
            gi = np.zeros(r.shape,dtype=complex)
            # plane wave at sphere idp : use local spherical decomposition
            if isinstance(idp,list):
                # print('incident field at p sphere')
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[idp[0]])
                for l in range(nmax):
                    fi += 1J**l*(2*l+1)*spu.jn(l,k*r_p)*spu.Pl(l,theta_p)
                fi *= np.exp(1J*k*self.d_p[idp])
            # Otherwise actual plane wave propagating along z
            else:
                # print('full incident field')
                fi = np.exp(1J*k*z)
                gi = 1J*k*np.cos(theta)*np.exp(1J*k*z)
            #remove incident field inside pth spheres
            for p in range(N):
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
                idx_i = r_p<self.ka
                fi[idx_i] = 0
                gi[idx_i] = 0

        #scattered fields
        if ftype in 'sta':
            fs = np.zeros(r.shape,dtype=complex)
            gs = np.zeros(r.shape,dtype=complex)
            #outgoing field
            for p in idp:
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
                idx_o = r_p>=self.ka     #;print(idx_o.shape)
                for l in range(nmax):
                    Yl = spe.sph_harm(0,l,0,theta_p)
                    fs[idx_o] += self.bp[p*nmax+l]*spu.hn1(l, k*r_p[idx_o])*Yl[idx_o]
                    gs[idx_o] += self.bp[p*nmax+l]*spu.hn1p(l,k*r_p[idx_o])*Yl[idx_o]
            #remove scattered field inside spheres
            for p in range(N):
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
                idx_i = r_p<self.ka
                fs[idx_i] = 0
                gs[idx_i] = 0
            #inside field
            for p in idp:
                r_p,theta_p,phi_p = spu.cart2sphere(x,y,z-self.d_p[p])
                idx_i = r_p<self.ka     #;print(idx_i.shape)
                for l in range(nmax):
                    Yl = spe.sph_harm(0,l,0,theta_p)
                    fs[idx_i] += self.ap[p*nmax+l]*spu.jn( l,     n_p*k*r_p[idx_i])*Yl[idx_i]
                    gs[idx_i] += self.ap[p*nmax+l]*n_p*spu.jnp( l,n_p*k*r_p[idx_i])*Yl[idx_i]

        #total field
        if Gopt:
            # print('     return Gradient');
            if   ftype=='t' :return gs+gi
            elif ftype=='s' :return gs
            elif ftype=='i' :return gi
            elif ftype=='a' :return gi,gs
        else:
            if   ftype=='t' :return fs+fi
            elif ftype=='s' :return fs
            elif ftype=='i' :return fi
            elif ftype=='a' :return fi,fs

    def get_ff(self,theta):
        k,d_p,nmax,N = self.k,self.d_p,self.nmax,self.N
        ct = np.cos(theta)

        f = np.zeros(theta.shape,dtype=complex)
        for l in range(nmax):
            Yl = spe.sph_harm(0,l,0,theta)
            for p in np.arange(N):
                f += (-1J)**(l+1)*Yl*self.bp[p*nmax+l]*np.exp(-1J*self.kd_p[p]*ct)
        return f
